#!/usr/bin/python
#
# Copyright (c) 2024 xuzhang3 (@xuzhang3), Fred-sun (@Fred-sun)
#
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_storageaccountmanagementpolicy
version_added: "2.4.0"
short_description: Manage storage account management policies
description:
    - Create, update or delete storage account management policies.
options:
    resource_group:
        description:
            - Name of the resource group to use.
        required: true
        type: str
        aliases:
            - resource_group_name
    storage_account_name:
        description:
            - Name of the storage account.
        type: str
        required: true
    rules:
        description:
            - The Storage Account ManagementPolicies Rules.
        type: list
        elements: dict
        suboptions:
            name:
                description:
                    - The name of the policy rule.
                    - A rule name can contain any combination of alpha numeric characters.
                type: str
                required: true
            type:
                description:
                    - The type of the policy rule.
                type: str
                required: true
                choices:
                    - Lifecycle
            enabled:
                description:
                    - Whether to enabled the rule
                type: bool
            definition:
                description:
                    - Whether to enabled the rule
                required: true
                type: dict
                suboptions:
                    actions:
                        description:
                            - An object that defines the action set.
                        type: dict
                        required: true
                        suboptions:
                            base_blob:
                                description:
                                    - The management policy action for base blob.
                                type: dict
                                suboptions:
                                    tier_to_cool:
                                        description:
                                            - The function to tier blobs to cool storage.
                                            - Support blobs currently at Hot tier.
                                        type: dict
                                        suboptions:
                                            days_after_modification_greater_than:
                                                description:
                                                    - Value indicating the age in days after last modification.
                                                type: float
                                            days_after_last_access_time_greater_than:
                                                description:
                                                    - This property can only be used in conjunction with last access time tracking policy.
                                                type: float
                                    tier_to_archive:
                                        description:
                                            - The function to tier blobs to archive storage.
                                            - Support blobs currently at Hot or Cool tier.
                                        type: dict
                                        suboptions:
                                            days_after_modification_greater_than:
                                                description:
                                                    - Value indicating the age in days after last modification.
                                                type: float
                                            days_after_last_access_time_greater_than:
                                                description:
                                                    - This property can only be used in conjunction with last access time tracking policy.
                                                type: float
                                    delete:
                                        description:
                                            - The function to delete the blob.
                                        type: dict
                                        suboptions:
                                            days_after_modification_greater_than:
                                                description:
                                                    - Value indicating the age in days after last modification.
                                                type: float
                                            days_after_last_access_time_greater_than:
                                                description:
                                                    - This property can only be used in conjunction with last access time tracking policy.
                                                type: float
                                    enable_auto_tier_to_hot_from_cool:
                                        description:
                                            - This property enables auto tiering of a blob from cool to hot on a blob access.
                                        type: bool
                            snapshot:
                                description:
                                    - The management policy action for snapshot.
                                type: dict
                                suboptions:
                                    tier_to_cool:
                                        description:
                                            - The function to tier blob snapshot to cool storage.
                                            - Support blob snapshot at Hot tier.
                                        type: dict
                                        suboptions:
                                            days_after_creation_greater_than:
                                                description:
                                                    - Value indicating the age in days after creation.
                                                type: float
                                                required: true
                                    tier_to_archive:
                                        description:
                                            - The function to tier blob snapshot to archive storage.
                                            - Support blob snapshot currently at Hot or Cool tier.
                                        type: dict
                                        suboptions:
                                            days_after_creation_greater_than:
                                                description:
                                                    - Value indicating the age in days after creation.
                                                type: float
                                                required: true
                                    delete:
                                        description:
                                            - The function to delete the blob snapshot.
                                        type: dict
                                        suboptions:
                                            days_after_creation_greater_than:
                                                description:
                                                    - Value indicating the age in days after creation.
                                                type: float
                                                required: true
                            version:
                                description:
                                    - The management policy action for version.
                                type: dict
                                suboptions:
                                    tier_to_cool:
                                        description:
                                            - The function to tier blob version to cool storage.
                                            - Support blob version currently at Hot tier.
                                        type: dict
                                        suboptions:
                                            days_after_creation_greater_than:
                                                description:
                                                    - Value indicating the age in days after creation.
                                                type: float
                                                required: true
                                    tier_to_archive:
                                        description:
                                            - The function to tier blob version to archive storage.
                                            - Support blob version currently at Hot or Cool tier.
                                        type: dict
                                        suboptions:
                                            days_after_creation_greater_than:
                                                description:
                                                    - Value indicating the age in days after creation.
                                                type: float
                                                required: true
                                    delete:
                                        description:
                                            - The function to delete the blob version.
                                        type: dict
                                        suboptions:
                                            days_after_creation_greater_than:
                                                description:
                                                    - Value indicating the age in days after creation.
                                                type: float
                                                required: true
                    filters:
                        description:
                            - An object that defines the filter set.
                        type: dict
                        suboptions:
                            prefix_match:
                                description:
                                    - An array of strings for prefixes to be match.
                                type: list
                                elements: str
                            blob_types:
                                description:
                                    - An array of predefined enum values.
                                    - Currently blockBlob supports all tiering and delete actions. Only delete actions are supported for C(appendBlob).
                                type: list
                                required: true
                                elements: str
                                choices:
                                    - blockBlob
                                    - appendBlob
                            blob_index_match:
                                description:
                                    - An array of blob index tag based filters, there can be at most 10 tag filters.
                                type: list
                                elements: dict
                                suboptions:
                                    name:
                                        description:
                                            - This is the filter tag name, it can have 1 - 128 characters.
                                        type: str
                                        required: true
                                    op:
                                        description:
                                            - This is the comparison operator which is used for object comparison and filtering.
                                            - Only C(==) (equality operator) is currently supported.
                                        type: str
                                        required: true
                                    value:
                                        description:
                                            - This is the filter tag value field used for tag based filtering.
                                            - It can have 0-256 characters.
                                        type: str
                                        required: true
    state:
        description:
            - State of the storage account managed policy. Use C(present) add or update the policy rule.
            - Use C(absent) to delete all policy rules.
        default: present
        type: str
        choices:
            - absent
            - present

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - xuzhang3 (@xuzhang3)
    - Fred-sun (@Fred-sun)
'''

EXAMPLES = '''
- name: Create storage account management policy with multi parameters
  azure_rm_storageaccountmanagementpolicy:
    resource_group: testRG
    storage_account_name: testaccount
    rules:
      - name: olcmtest5
        type: Lifecycle
        enabled: false
        definition:
          actions:
            base_blob:
              enable_auto_tier_to_hot_from_cool: true
              delete:
                days_after_modification_greater_than: 33
                days_after_last_access_time_greater_than: 33
              tier_to_cool:
                days_after_modification_greater_than: 33
                days_after_last_access_time_greater_than: 33
              tier_to_archive:
                days_after_modification_greater_than: 33
                days_after_last_access_time_greater_than: 33
            snapshot:
              tier_to_cool:
                days_after_creation_greater_than: 33
              tier_to_archive:
                days_after_creation_greater_than: 33
              delete:
                days_after_creation_greater_than: 33
            version:
              tier_to_archive:
                days_after_creation_greater_than: 33
              tier_to_cool:
                days_after_creation_greater_than: 33
              delete:
                days_after_creation_greater_than: 33
          filters:
            prefix_match:
              - olcmtestcontainer2
            blob_types:
              - blockBlob
              - appendBlob
            blob_index_match:
              - name: tags3
                op: '=='
                value: value3

- name: Delete management policy rules
  azure_rm_storageaccountmanagementpolicy:
    resource_group: "{{ resource_group }}"
    storage_account_name: "st{{ rpfx }}"
    state: absent
'''


RETURN = '''
state:
    description:
        - The data policy rules associated with the specified storage account.
    returned: always
    type: complex
    contains:
        id:
            description:
                - The data policy's ID.
            returned: always
            type: str
            sample: "/subscriptions/xxx-xxx/resourceGroups/testRG/providers/Microsoft.Storage/storageAccounts/sttest/managementPolicies/default"
        resource_group:
            description:
                - The resource group name.
            returned: always
            type: str
            sample: testRG
        storage_account_name:
            description:
                - The storage account name.
            returned: always
            type: str
            sample: teststname
        type:
            description:
                - The type of the resource.
            returned: always
            type: str
            sample: "Microsoft.Storage/storageAccounts/managementPolicies"
        last_modified_time:
            description:
                - Returns the date and time the ManagementPolicies was last modified.
            returned: always
            type: str
            sample: "2024-04-12T11:40:10.376465+00:00"
        name:
            description:
                - The name of the resource.
            returned: always
            type: str
            sample: DefaultManagementPolicy
        policy:
            description:
                - The Storage Account ManagementPolicy.
            returned: always
            type: complex
            contains:
                rules:
                    description:
                        - The Storage Account ManagementPolicies Rules.
                    type: list
                    returned: always
                    sample: [
                    {
                        "definition": {
                            "actions": {
                                "base_blob": {
                                    "delete": {
                                        "days_after_last_access_time_greater_than": 33.0,
                                        "days_after_modification_greater_than": 33.0
                                    },
                                    "enable_auto_tier_to_hot_from_cool": true,
                                    "tier_to_archive": {
                                        "days_after_last_access_time_greater_than": 33.0,
                                        "days_after_modification_greater_than": 33.0
                                    },
                                    "tier_to_cool": {
                                        "days_after_last_access_time_greater_than": 33.0,
                                        "days_after_modification_greater_than": 33.0
                                    }
                                },
                                "snapshot": {
                                    "delete": {
                                        "days_after_creation_greater_than": 33.0
                                    },
                                    "tier_to_archive": {
                                        "days_after_creation_greater_than": 33.0
                                    },
                                    "tier_to_cool": {
                                        "days_after_creation_greater_than": 33.0
                                    }
                                },
                                "version": {
                                    "delete": {
                                        "days_after_creation_greater_than": 33.0
                                    },
                                    "tier_to_archive": {
                                        "days_after_creation_greater_than": 33.0
                                    },
                                    "tier_to_cool": {
                                        "days_after_creation_greater_than": 33.0
                                    }
                                }
                            },
                            "filters": {
                                "blob_index_match": [
                                    {
                                        "name": "tags3",
                                        "op": "==",
                                        "value": "value3"
                                    }
                                ],
                                "blob_types": [
                                    "blockBlob",
                                    "appendBlob"
                                ],
                                "prefix_match": [
                                    "olcmtestcontainer2"
                                ]
                            }
                        },
                        "enabled": false,
                        "name": "olcmtest5",
                        "type": "Lifecycle"
                    }
                ]
'''


from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ext import AzureRMModuleBaseExt
from ansible.module_utils.common.dict_transformations import snake_dict_to_camel_dict
try:
    from azure.core.exceptions import ResourceNotFoundError
    from azure.core.serialization import as_attribute_dict
except Exception:
    # This is handled in azure_rm_common
    pass


class AzureRMStorageAccountManagementPolicy(AzureRMModuleBaseExt):

    def __init__(self):

        self.module_arg_spec = dict(
            resource_group=dict(required=True, type='str', aliases=['resource_group_name']),
            storage_account_name=dict(type='str', required=True),
            state=dict(default='present', choices=['present', 'absent']),
            rules=dict(
                type='list',
                elements='dict',
                options=dict(
                    enabled=dict(type='bool'),
                    name=dict(type='str', required=True),
                    type=dict(type='str', required=True, choices=['Lifecycle']),
                    definition=dict(
                        type='dict',
                        required=True,
                        options=dict(
                            actions=dict(
                                type='dict',
                                required=True,
                                options=dict(
                                    base_blob=dict(
                                        type='dict',
                                        options=dict(
                                            tier_to_cool=dict(
                                                type='dict',
                                                options=dict(
                                                    days_after_modification_greater_than=dict(type='float'),
                                                    days_after_last_access_time_greater_than=dict(type='float')
                                                )
                                            ),
                                            tier_to_archive=dict(
                                                type='dict',
                                                options=dict(
                                                    days_after_modification_greater_than=dict(type='float'),
                                                    days_after_last_access_time_greater_than=dict(type='float')
                                                )
                                            ),
                                            delete=dict(
                                                type='dict',
                                                options=dict(
                                                    days_after_modification_greater_than=dict(type='float'),
                                                    days_after_last_access_time_greater_than=dict(type='float')
                                                )
                                            ),
                                            enable_auto_tier_to_hot_from_cool=dict(type='bool')
                                        )
                                    ),
                                    snapshot=dict(
                                        type='dict',
                                        options=dict(
                                            tier_to_cool=dict(
                                                type='dict',
                                                options=dict(
                                                    days_after_creation_greater_than=dict(type='float', required=True)
                                                )
                                            ),
                                            tier_to_archive=dict(
                                                type='dict',
                                                options=dict(
                                                    days_after_creation_greater_than=dict(type='float', required=True)
                                                )
                                            ),
                                            delete=dict(
                                                type='dict',
                                                options=dict(
                                                    days_after_creation_greater_than=dict(type='float', required=True)
                                                )
                                            )
                                        )
                                    ),
                                    version=dict(
                                        type='dict',
                                        options=dict(
                                            tier_to_cool=dict(
                                                type='dict',
                                                options=dict(
                                                    days_after_creation_greater_than=dict(
                                                        type='float',
                                                        required=True
                                                    )
                                                )
                                            ),
                                            tier_to_archive=dict(
                                                type='dict',
                                                options=dict(
                                                    days_after_creation_greater_than=dict(
                                                        type='float',
                                                        required=True
                                                    )
                                                )
                                            ),
                                            delete=dict(
                                                type='dict',
                                                options=dict(
                                                    days_after_creation_greater_than=dict(
                                                        type='float',
                                                        required=True
                                                    )
                                                )
                                            )
                                        )
                                    )
                                )
                            ),
                            filters=dict(
                                type='dict',
                                options=dict(
                                    prefix_match=dict(type='list', elements='str'),
                                    blob_types=dict(type='list', elements='str', choices=['blockBlob', 'appendBlob'], required=True),
                                    blob_index_match=dict(
                                        type='list',
                                        elements='dict',
                                        options=dict(
                                            name=dict(type='str', required=True),
                                            op=dict(type='str', required=True),
                                            value=dict(type='str', required=True)
                                        )
                                    )
                                )
                            )
                        )
                    )
                )
            )
        )

        self.results = dict(
            changed=False,
            state=dict()
        )

        self.resource_group = None
        self.storage_account_name = None
        self.state = None
        self.rules = []

        super(AzureRMStorageAccountManagementPolicy, self).__init__(self.module_arg_spec,
                                                                    supports_tags=False,
                                                                    supports_check_mode=True)

    def exec_module(self, **kwargs):

        for key in list(self.module_arg_spec.keys()):
            setattr(self, key, kwargs[key])

        managed_policy = self.get_management_policy()
        changed = False

        if self.state == 'present':
            if managed_policy is not None:
                rules = []
                for item in managed_policy['policy']['rules']:
                    rules.append(item)
                rules_name = [item['name'] for item in rules]
                for item in self.rules:
                    if item['name'] in rules_name:
                        for tt in managed_policy['policy']['rules']:
                            if item['name'] == tt['name']:
                                old_item = tt
                                if not self.default_compare({}, item, old_item, '', dict(compare=[])):
                                    rules.remove(old_item)
                                    rules.append(item)
                                    changed = True
                    else:
                        rules.append(item)
                        changed = True
                if changed and not self.check_mode:
                    self.create_or_update_management_policy(rules)
            else:
                changed = True
                if not self.check_mode:
                    self.create_or_update_management_policy(self.rules)
        else:
            if managed_policy is not None:
                changed = True
                if not self.check_mode:
                    self.delete_management_policy()

        self.results['state'] = self.get_management_policy()
        self.results['changed'] = changed

        return self.results

    def get_management_policy(self):
        self.log('Get info for storage account management policy')

        response = None
        try:
            response = self.storage_client.management_policies.get(self.resource_group, self.storage_account_name, 'default')
        except ResourceNotFoundError as ec:
            self.log("Failed to obtain the storage acount management policy, detail as {0}".format(ec))
            return None
        return self.format_to_dict(response)

    def create_or_update_management_policy(self, rules):
        self.log("Creating or updating storage account mangement policy")

        try:
            self.storage_client.management_policies.create_or_update(
                resource_group_name=self.resource_group,
                account_name=self.storage_account_name,
                management_policy_name='default',
                properties={'properties': {'policy': {'rules': [snake_dict_to_camel_dict(rule) for rule in rules]}}})
        except Exception as e:
            self.log('Error creating or updating storage account management policy.')
            self.fail("Failed to create or updating storage account management policy: {0}".format(str(e)))
        return self.get_management_policy()

    def delete_management_policy(self):
        try:
            self.storage_client.management_policies.delete(self.resource_group, self.storage_account_name, 'default')
        except Exception as e:
            self.fail("Failed to delete the storage account management policy: {0}".format(str(e)))

    def format_to_dict(self, obj):
        result = dict()
        result['id'] = obj.id
        result['resource_group'] = self.resource_group
        result['storage_account_name'] = self.storage_account_name
        result['name'] = obj.name
        result['type'] = obj.type
        result['last_modified_time'] = obj.last_modified_time
        result['policy'] = dict(rules=[])
        if obj.policy is not None:
            result['policy'] = as_attribute_dict(obj.policy)

        return result


def main():
    AzureRMStorageAccountManagementPolicy()


if __name__ == '__main__':
    main()



#
# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
The nxos_acl_interfaces class
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to it's desired end-state is
created
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.cfg.base import (
    ConfigBase,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    remove_empties,
    to_list,
)

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.facts.facts import Facts
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.utils.utils import (
    get_interface_type,
    normalize_interface,
    search_obj_in_list,
)


class Acl_interfaces(ConfigBase):
    """
    The nxos_acl_interfaces class
    """

    gather_subset = ["!all", "!min"]

    gather_network_resources = ["acl_interfaces"]

    def __init__(self, module):
        super(Acl_interfaces, self).__init__(module)

    def get_acl_interfaces_facts(self, data=None):
        """Get the 'facts' (the current configuration)

        :rtype: A dictionary
        :returns: The current configuration as a dictionary
        """
        facts, _warnings = Facts(self._module).get_facts(
            self.gather_subset,
            self.gather_network_resources,
            data=data,
        )
        acl_interfaces_facts = facts["ansible_network_resources"].get("acl_interfaces")
        if not acl_interfaces_facts:
            return []
        return acl_interfaces_facts

    def edit_config(self, commands):
        """Wrapper method for `_connection.edit_config()`
        This exists solely to allow the unit test framework to mock device connection calls.
        """
        return self._connection.edit_config(commands)

    def execute_module(self):
        """Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        result = {"changed": False}
        warnings = list()
        commands = list()
        self.state = self._module.params["state"]
        action_states = ["merged", "replaced", "deleted", "overridden"]

        if self.state == "gathered":
            result["gathered"] = self.get_acl_interfaces_facts()
        elif self.state == "rendered":
            result["rendered"] = self.set_config({})
            # no need to fetch facts for rendered
        elif self.state == "parsed":
            running_config = self._module.params["running_config"]
            if not running_config:
                self._module.fail_json(
                    msg="value of running_config parameter must not be empty for state parsed",
                )
            result["parsed"] = self.set_config({})
            # no need to fetch facts for parsed
        else:
            existing_acl_interfaces_facts = self.get_acl_interfaces_facts()
            commands.extend(self.set_config(existing_acl_interfaces_facts))
            if commands and self.state in action_states:
                if not self._module.check_mode:
                    self._connection.edit_config(commands)
                result["changed"] = True
                result["before"] = existing_acl_interfaces_facts
                result["commands"] = commands

            changed_acl_interfaces_facts = self.get_acl_interfaces_facts()
            if result["changed"]:
                result["after"] = changed_acl_interfaces_facts
        result["warnings"] = warnings
        return result

    def set_config(self, existing_acl_interfaces_facts):
        """Collect the configuration from the args passed to the module,
            collect the current configuration (as a dict from facts)

        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        config = self._module.params["config"]
        want = []
        if config:
            for w in config:
                if get_interface_type(w["name"]) == "loopback":
                    self._module.fail_json(
                        msg="This module works with ethernet, management or port-channel.",
                    )
                w.update({"name": normalize_interface(w["name"])})
                want.append(remove_empties(w))
        have = existing_acl_interfaces_facts
        resp = self.set_state(want, have)
        return to_list(resp)

    def set_state(self, want, have):
        """Select the appropriate function based on the state provided

        :param want: the desired configuration as a dictionary
        :param have: the current configuration as a dictionary
        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        if self.state in ("overridden", "merged", "replaced", "rendered") and not want:
            self._module.fail_json(
                msg="value of config parameter must not be empty for state {0}".format(
                    self.state,
                ),
            )

        commands = []
        if self.state == "overridden":
            commands = self._state_overridden(want, have)
        elif self.state == "deleted":
            commands = self._state_deleted(want, have)
        elif self.state == "rendered":
            commands = self._state_rendered(want)
        elif self.state == "parsed":
            want = self._module.params["running_config"]
            commands = self._state_parsed(want)
        else:
            for w in want:
                if self.state == "merged":
                    commands.extend(self._state_merged(w, have))
                elif self.state == "replaced":
                    commands.extend(self._state_replaced(w, have))
        return commands

    def _state_parsed(self, want):
        return self.get_acl_interfaces_facts(want)

    def _state_rendered(self, want):
        commands = []
        for w in want:
            commands.extend(self.set_commands(w, {}))
        return commands

    def _state_replaced(self, want, have):
        """The command generator when state is replaced

        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        new_commands = []
        del_dict = {"name": want["name"], "access_groups": []}
        obj_in_have = search_obj_in_list(want["name"], have, "name")
        if obj_in_have != want:
            commands = []
            if obj_in_have and "access_groups" in obj_in_have.keys():
                for ag in obj_in_have["access_groups"]:
                    want_afi = []
                    if want.get("access_groups"):
                        want_afi = search_obj_in_list(
                            ag["afi"],
                            want["access_groups"],
                            "afi",
                        )
                    if not want_afi:
                        # whatever in have is not in want
                        del_dict["access_groups"].append(ag)
                    else:
                        del_acl = []
                        for acl in ag["acls"]:
                            if want_afi.get("acls"):
                                if acl not in want_afi["acls"]:
                                    del_acl.append(acl)
                            else:
                                del_acl.append(acl)
                        afi = want_afi["afi"]
                        del_dict["access_groups"].append({"afi": afi, "acls": del_acl})

            commands.extend(self._state_deleted([del_dict], have))
            commands.extend(self._state_merged(want, have))
            new_commands.append(commands[0])
            commands = [commands[i] for i in range(1, len(commands)) if commands[i] != commands[0]]
            new_commands.extend(commands)
        return new_commands

    def _state_overridden(self, want, have):
        """The command generator when state is overridden

        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        commands = []
        want_intf = [w["name"] for w in want]
        for h in have:
            if h["name"] not in want_intf:
                commands.extend(self._state_deleted([h], have))
        for w in want:
            commands.extend(self._state_replaced(w, have))
        return commands

    def _state_merged(self, want, have):
        """The command generator when state is merged

        :rtype: A list
        :returns: the commands necessary to merge the provided into
                  the current configuration
        """
        return self.set_commands(want, have)

    def set_commands(self, want, have, deleted=False):
        commands = []
        have_name = search_obj_in_list(want["name"], have, "name")
        if have_name and have_name.get("access_groups"):
            if want.get("access_groups"):
                for w_afi in want["access_groups"]:
                    ip = "ipv6"
                    if w_afi["afi"] == "ipv4":
                        ip = "ip"
                    have_afi = search_obj_in_list(
                        w_afi["afi"],
                        have_name["access_groups"],
                        "afi",
                    )
                    if have_afi:
                        new_acls = []
                        if deleted:
                            if w_afi.get("acls") and have_afi.get("acls"):
                                new_acls = [
                                    acl for acl in w_afi.get("acls") if acl in have_afi.get("acls")
                                ]
                            elif "acls" not in w_afi.keys():
                                new_acls = have_afi.get("acls")
                        else:
                            if w_afi.get("acls"):
                                new_acls = [
                                    acl for acl in w_afi["acls"] if acl not in have_afi["acls"]
                                ]
                        commands.extend(self.process_acl(new_acls, ip, deleted))
                    else:
                        if not deleted:
                            if w_afi.get("acls"):
                                commands.extend(self.process_acl(w_afi["acls"], ip))
            else:
                # only name is given to delete
                if deleted and "access_groups" in have_name.keys():
                    commands.extend(self.process_access_group(have_name, True))
        else:
            if not deleted:  # and 'access_groups' in have_name.keys():
                commands.extend(self.process_access_group(want))

        if len(commands) > 0:
            commands.insert(0, "interface " + want["name"])
        return commands

    def process_access_group(self, item, deleted=False):
        commands = []
        for ag in item["access_groups"]:
            ip = "ipv6"
            if ag["afi"] == "ipv4":
                ip = "ip"
            if ag.get("acls"):
                commands.extend(self.process_acl(ag["acls"], ip, deleted))
        return commands

    def process_acl(self, acls, ip, deleted=False):
        commands = []
        no = ""
        if deleted:
            no = "no "
        for acl in acls:
            port = ""
            if acl.get("port"):
                port = " port"
            ag = " access-group "
            if ip == "ipv6":
                ag = " traffic-filter "
            commands.append(no + ip + port + ag + acl["name"] + " " + acl["direction"])
        return commands

    def _state_deleted(self, main_want, have):
        """The command generator when state is deleted

        :rtype: A list
        :returns: the commands necessary to remove the current configuration
                  of the provided objects
        """
        commands = []
        if main_want:
            if self.state == "deleted":
                for w in main_want:
                    h = search_obj_in_list(w["name"], have, "name") or {}
                    commands.extend(self.set_commands(h, have, deleted=True))
            else:
                for want in main_want:
                    commands.extend(self.set_commands(want, have, deleted=True))
        else:
            for h in have:
                commands.extend(self.set_commands(h, have, deleted=True))

        return commands



# Copyright (c) 2021 Felix Fontein
# Copyright (c) 2020 Markus Bergholz <markuman+spambelongstogoogle@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
name: hosttech_dns_records

short_description: Create inventory from Hosttech DNS records

version_added: 2.0.0

description:
  - This plugin allows to create an inventory from Hosttech DNS records.
  - 'For Ansible to be able to identify a YAML file as an inventory for this plugin, the inventory file must contain C(plugin:
    community.dns.hosttech_dns_records) and its filename must end with C(hosttech_dns.yaml) or C(hosttech_dns.yml).'
options:
  plugin:
    description: The name of this plugin. Should always be set to V(community.dns.hosttech_dns_records) for this plugin to
      recognize it as its own.
    required: true
    choices:
      - community.dns.hosttech_dns_records
    type: str

    # We need to overwrite zone_id to be of type string, otherwise templating cannot be passed in
  zone_id:
    type: raw
    # If there would not be ansible-base 2.10, this should be string instead. ansible-base will
    # not accept an integer for type=string options, whence type=string breaks backwards
    # compatibility with previous type=int...
    #   type: string

  filters:
    version_added: 3.0.0

extends_documentation_fragment:
  - community.dns._hosttech
  - community.dns._hosttech.plugin
  - community.dns._hosttech.record_type_choices_records_inventory
  - community.dns._hosttech.record_type_seealso
  - community.dns._hosttech.zone_id_type
  - community.dns._inventory_records
  - community.dns._options.record_transformation
  - community.library_inventory_filtering_v1.inventory_filter

notes:
  - The provider-specific O(hosttech_username), O(hosttech_password), and O(hosttech_token) options can be templated.
author:
  - Markus Bergholz (@markuman) <markuman+spambelongstogoogle@gmail.com>
  - Felix Fontein (@felixfontein)

seealso:
  - module: community.dns.hosttech_dns_record_set_info
  - module: community.dns.hosttech_dns_record_info
"""

EXAMPLES = r"""
# filename must end with hosttech_dns.yaml or hosttech_dns.yml

plugin: community.dns.hosttech_dns_records
zone_name: domain.ch
simple_filters:
  type:
    - AAAA
filters:
  - include: >-
      '*.' not in inventory_hostname
  - exclude: true

# You can also configure the token by putting secret value into this file,
# but this is discouraged. Use a lookup like below, or leave it away and
# set it with the ANSIBLE_HOSTTECH_DNS_TOKEN environment variable.
hosttech_token: >-
  {{ (lookup('community.sops.sops', 'keys/hosttech.sops.yml') | from_yaml).hosttech_dns_token }}
"""

import typing as t

from ansible_collections.community.dns.plugins.module_utils._hosttech.api import (
    create_hosttech_api,
    create_hosttech_provider_information,
)
from ansible_collections.community.dns.plugins.module_utils._http import OpenURLHelper
from ansible_collections.community.dns.plugins.plugin_utils._inventory.records import (
    RecordsInventoryModule,
)
from ansible_collections.community.dns.plugins.plugin_utils._templated_options import (
    TemplatedOptionProvider,
)

if t.TYPE_CHECKING:
    from ..module_utils._provider import ProviderInformation  # pragma: no cover
    from ..module_utils._zone_record_api import ZoneRecordAPI  # pragma: no cover
    from ..module_utils._zone_record_set_api import ZoneRecordSetAPI  # pragma: no cover


class InventoryModule(RecordsInventoryModule):
    NAME = "community.dns.hosttech_dns_records"
    VALID_ENDINGS = ("hosttech_dns.yaml", "hosttech_dns.yml")

    def setup_api(self) -> tuple[ProviderInformation, ZoneRecordAPI | ZoneRecordSetAPI]:
        provider_information = create_hosttech_provider_information()
        return provider_information, create_hosttech_api(
            TemplatedOptionProvider(self, self.templar), OpenURLHelper()
        )



# (c) 2022 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
#

from __future__ import absolute_import, division, print_function


__metaclass__ = type

from functools import wraps

from ansible.plugins import AnsiblePlugin


def ensure_connect(func):
    @wraps(func)
    def wrapped(self, *args, **kwargs):
        if not self._connection._connected:
            self._connection._connect()
        return func(self, *args, **kwargs)

    return wrapped


class GrpcBase(AnsiblePlugin):
    """
    A base class for implementing gRPC abstraction layer
    """

    __rpc__ = ["channel", "get_config", "edit_config", "get"]

    def __init__(self, connection):
        super(GrpcBase, self).__init__()
        self._connection = connection

    @property
    @ensure_connect
    def channel(self):
        return self._connection._channel

    def get_config(self, section=None):
        """
        Retrieve all or part of a specified configuration
        (by default entire configuration is retrieved).
        :param section: This argument specifies the portion of the configuration data to retrieve
        :return: Returns the response received from gRPC server from target host in string format
        """
        pass

    def get(self, section=None):
        """
        Retrieve device state information.
        :param section: This argument specifies the portion of the state data to retrieve
                       (by default entire state data is retrieved)
        :return: Returns the json string as a response
        """
        pass

    def edit_config(self, config=None, action=None):
        """
        Loads all or part of the specified *config* to the configuration datastore.
        :param config: The configuration that needs to be push on target host
        :param action: The action to be performed on the configuration datastore for example: 'merge',
                       'replace', 'delete' etc.
        :return: Returns the response received from gRPC server from target host in string format
        """
        pass



#!/usr/bin/python
#
# Copyright (c) 2021 Andrii Bilorus, <andrii.bilorus@gmail.com>
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_aksupgrade_info

version_added: "1.4.0"

short_description: Get the upgrade versions available for a AKS instance

description:
    - Get the upgrade versions available for a managed Azure Container Service (AKS) instance.

options:
    resource_group:
        description:
            - Name of a resource group where the managed Azure Container Services (AKS) exists.
        required: true
        type: str
    name:
        description:
            - Name of the managed Azure Container Services (AKS) instance.
        required: true
        type: str

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Andrii Bilorus (@ewscat)
'''

EXAMPLES = '''
- name: Get available upgrade versions for AKS instance
  azure_rm_aksupgrade_info:
    name: myAKS
    resource_group: myResourceGroup
  register: myAKSupgrades
'''

RETURN = '''
azure_aks_upgrades:
    description: Supported AKS instance versions for upgrade by agent pools and control plane.
    returned: always
    type: complex
    contains:
        agent_pool_profiles:
            description: Available upgrade versions for agent pools
            returned: always
            type: complex
            contains:
                upgrades:
                    description: List of orchestrator types and versions available for upgrade.
                    type: complex
                    contains:
                        is_preview:
                            description: Is the version available in preview
                            type: bool
                        kubernetes_version:
                            description: Kubernetes version
                            type: str
                            sample: "1.19.3"
                os_type:
                    description: Operating system type
                    type: str
                    sample: "Linux"
                name:
                    description: Pool name
                    type: str
                    sample: "my_pool"
                kubernetes_version:
                    description: Current kubernetes version
                    type: str
                    sample: "1.18.1"
        control_plane_profile:
            description: Available upgrade versions for control plane
            returned: always
            type: complex
            contains:
                upgrades:
                    description: List of orchestrator types and versions available for upgrade.
                    type: complex
                    contains:
                        is_preview:
                            description: Is the version available in preview
                            type: bool
                        kubernetes_version:
                            description: Kubernetes version
                            type: str
                            sample: "1.19.3"
                os_type:
                    description: Operating system type
                    type: str
                    sample: "Linux"
                name:
                    description: Pool name
                    type: str
                    sample: "my_pool"
                kubernetes_version:
                    description: Current kubernetes version
                    type: str
                    sample: "1.18.1"
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase

try:
    from azure.core.exceptions import ResourceNotFoundError
except Exception:
    # handled in azure_rm_common
    pass


class AzureRMAKSUpgrade(AzureRMModuleBase):
    '''
    Utility class to get Azure Kubernetes Service upgrades
    '''

    def __init__(self):

        self.module_args = dict(
            name=dict(type='str', required=True),
            resource_group=dict(type='str', required=True)
        )

        self.results = dict(
            changed=False,
            azure_aks_upgrades=[]
        )

        self.name = None
        self.resource_group = None

        super(AzureRMAKSUpgrade, self).__init__(
            derived_arg_spec=self.module_args,
            supports_check_mode=True,
            supports_tags=False,
            facts_module=True
        )

    def exec_module(self, **kwargs):
        for key in self.module_args:
            setattr(self, key, kwargs[key])

        self.results['azure_aks_upgrades'] = self.get_upgrades(self.name, self.resource_group)

        return self.results

    def get_upgrades(self, name, resource_group):
        '''
        Get supported upgrade version for AKS
        :param: name: str with name of AKS cluster instance
        :param: resource_group: str with resource group containing AKS instance
        :return: dict with available versions for pool profiles and control plane
        '''
        cluster = None
        upgrade_profiles = None

        self.log('Get properties for {0}'.format(self.name))
        try:
            cluster = self.managedcluster_client.managed_clusters.get(resource_group_name=resource_group, resource_name=name)
        except ResourceNotFoundError as err:
            self.fail('Error when getting AKS cluster information for {0} : {1}'.format(self.name, err.message or str(err)))

        self.log('Get available upgrade versions for {0}'.format(self.name))
        try:
            upgrade_profiles = self.managedcluster_client.managed_clusters.get_upgrade_profile(resource_group_name=resource_group,
                                                                                               resource_name=name)
        except ResourceNotFoundError as err:
            self.fail('Error when getting upgrade versions for {0} : {1}'.format(self.name, err.message or str(err)))

        return dict(
            agent_pool_profiles=[self.parse_profile(profile)
                                 if profile.upgrades else self.default_profile(cluster)
                                 for profile in upgrade_profiles.agent_pool_profiles]
            if upgrade_profiles.agent_pool_profiles else None,
            control_plane_profile=self.parse_profile(upgrade_profiles.control_plane_profile)
            if upgrade_profiles.control_plane_profile.upgrades
            else self.default_profile(cluster)
        )

    def default_profile(self, cluster):
        '''
        Used when upgrade profile returned by Azure in None
        (i.e. when the cluster runs latest version)
        :param: cluster: ManagedCluster with AKS instance information
        :return: dict containing upgrade profile with current cluster version
        '''
        return dict(
            upgrades=None,
            kubernetes_version=cluster.kubernetes_version,
            name=None,
            os_type=None
        )

    def parse_profile(self, profile):
        '''
        Transform cluster profile object to dict
        :param: profile: ManagedClusterUpgradeProfile with AKS upgrade profile info
        :return: dict with upgrade profiles
        '''
        return dict(
            upgrades=[dict(
                is_preview=upgrade.is_preview,
                kubernetes_version=upgrade.kubernetes_version
            ) for upgrade in profile.upgrades],
            kubernetes_version=profile.kubernetes_version,
            name=profile.name,
            os_type=profile.os_type
        )


def main():
    '''
    Main module execution code path
    '''
    AzureRMAKSUpgrade()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The Logging_global parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


def tmplt_host(verb):
    cmd = "logging host"
    changed = True
    if verb.get("transport"):
        changed = False
    if verb:
        if verb.get("host"):
            cmd += " {hostname}".format(hostname=verb["host"])
        if verb.get("ipv6"):
            cmd += " ipv6 {ipv6}".format(ipv6=verb["ipv6"])
        if verb.get("vrf"):
            cmd += " vrf {vrf}".format(vrf=verb["vrf"])
        if verb.get("filtered"):
            cmd += " {filtered}".format(filtered="filtered")
            changed = True
        if verb.get("xml"):
            cmd += " {xml}".format(xml="xml")
            changed = True
        if verb.get("session_id"):
            session_id = verb.get("session_id")
            changed = True
            if session_id.get("text"):
                cmd += " session-id string {text}".format(text=session_id["text"])
            elif session_id.get("tag"):
                cmd += " session-id {tag}".format(tag=session_id["tag"])
        if verb.get("stream"):
            cmd += " stream {stream}".format(stream=verb["stream"])
            changed = True
        if verb.get("sequence_num_session"):
            cmd += " {sequence_num_session}".format(sequence_num_session="sequence-num-session")
            changed = True
        if verb.get("discriminator"):
            cmd += " discriminator {discriminator}".format(discriminator=verb["discriminator"])
            changed = True
    if not changed:
        cmd = None
    return cmd


def tmplt_host_transport(verb):
    cmd = "logging host"

    if verb.get("host"):
        cmd += " {hostname}".format(hostname=verb["host"])
    if verb.get("ipv6"):
        cmd += " ipv6 {ipv6}".format(ipv6=verb["ipv6"])
    if verb.get("vrf"):
        cmd += " vrf {vrf}".format(vrf=verb["vrf"])
    if verb.get("transport"):
        transport_type = verb.get("transport")
        prot = None
        if transport_type.get("udp"):
            cmd += " transport {prot}".format(prot="udp")
            prot = "udp"
        elif transport_type.get("tcp"):
            cmd += " transport {prot}".format(prot="tcp")
            prot = "tcp"

        if prot:
            verb = transport_type.get(prot)

            if verb.get("port"):
                cmd += " port {port}".format(port=verb["port"])
            if verb.get("audit"):
                cmd += " {audit}".format(audit="audit")
            if verb.get("xml"):
                cmd += " {xml}".format(xml="xml")
            if verb.get("filtered"):
                cmd += " {filtered}".format(filtered="filtered")
            if verb.get("discriminator"):
                cmd += " discriminator {discriminator}".format(discriminator=verb["discriminator"])
            if verb.get("stream"):
                cmd += " stream {stream}".format(stream=verb["stream"])
            if verb.get("session_id"):
                session_id = verb.get("session_id")
                if session_id.get("text"):
                    cmd += " session-id string {text}".format(text=session_id["text"])
                elif session_id.get("tag"):
                    cmd += " session-id {tag}".format(tag=session_id["tag"])
            if verb.get("sequence_num_session"):
                cmd += " {sequence_num_session}".format(sequence_num_session="sequence-num-session")
    return cmd


def tmplt_host_del(verb):
    cmd = "logging host"
    if verb.get("host"):
        cmd += " {hostname}".format(hostname=verb["host"])
    if verb.get("ipv6"):
        cmd += " ipv6 {ipv6}".format(ipv6=verb["ipv6"])
    if verb.get("transport"):
        cmd = None
    return cmd


def tmplt_host_transport_del(verb):
    cmd = "logging host"
    if verb.get("host"):
        cmd += " {hostname}".format(hostname=verb["host"])
    if verb.get("ipv6"):
        cmd += " ipv6 {ipv6}".format(ipv6=verb["ipv6"])
    return cmd


def tmplt_buffered(config_data):
    return tmplt_common(config_data.get("buffered"), "logging buffered")


def tmplt_console(config_data):
    return tmplt_common(config_data.get("console"), "logging console")


def tmplt_monitor(config_data):
    return tmplt_common(config_data.get("monitor"), "logging monitor")


def tmplt_origin_id(config_data):
    return tmplt_common(config_data.get("origin_id"), "logging origin-id")


def tmplt_logging_on(config_data):
    cmd = "logging on"
    if config_data.get("logging_on") == "disable":
        cmd = "no " + cmd
    return cmd


def tmplt_queue_limit(config_data):
    return tmplt_common(config_data.get("queue_limit"), "logging queue-limit")


def tmplt_rate_limit(config_data):
    return tmplt_common(config_data.get("rate_limit"), "logging rate-limit")


def tmplt_reload(config_data):
    return tmplt_common(config_data.get("reload"), "logging reload")


def tmplt_message_counter(verb):
    cmd = "logging message-counter"

    if verb.get("message_counter"):
        cmd += " {message_counter}".format(message_counter=verb["message_counter"])
    return cmd


def tmplt_filter(verb):
    cmd = "logging filter"

    if verb.get("url"):
        cmd += " {url}".format(url=verb["url"])
    if verb.get("order"):
        cmd += " {order}".format(order=verb["order"])
    if verb.get("args"):
        cmd += " args {args}".format(args=verb["args"])
    return cmd


def tmplt_source_interface(verb):
    cmd = "logging source-interface"

    if verb.get("interface"):
        cmd += " {interface}".format(interface=verb["interface"])
    if verb.get("vrf"):
        cmd += " vrf {vrf}".format(vrf=verb["vrf"])
    return cmd


def tmplt_common(verb, cmd):
    if verb:
        if verb.get("all"):
            cmd += " {all}".format(all="all")
        if verb.get("console"):
            cmd += " {console}".format(console="console")
        if verb.get("message_limit"):
            cmd += " message-limit {message_limit}".format(message_limit=verb["message_limit"])
        if verb.get("discriminator"):
            cmd += " discriminator {discriminator}".format(discriminator=verb.get("discriminator"))
        if verb.get("filtered"):
            cmd += " {filtered}".format(filtered="filtered")
        if verb.get("xml"):
            cmd += " {xml}".format(xml="xml")
        if verb.get("size"):
            cmd += " {size}".format(size=verb["size"])
        if verb.get("severity"):
            cmd += " {severity}".format(severity=verb["severity"])
        if verb.get("except_severity"):
            cmd += " except {exceptSev}".format(exceptSev=verb["except_severity"])
        if verb.get("tag"):
            cmd += " {tag}".format(tag=verb["tag"])
        if verb.get("text"):
            cmd += " string {tag}".format(tag=verb["text"])
        if verb.get("esm"):
            cmd += " esm {tag}".format(tag=verb["esm"])
        if verb.get("trap"):
            cmd += " trap {tag}".format(tag=verb["trap"])
    return cmd


def tmplt_persistent(config_data):
    cmd = "logging persistent"
    verb = config_data.get("persistent")

    if verb.get("url"):
        cmd += " url {url}".format(url=verb["url"])
    if verb.get("size"):
        cmd += " size {size}".format(size=verb["size"])
    if verb.get("filesize"):
        cmd += " filesize {filesize}".format(filesize=verb["filesize"])
    if verb.get("batch"):
        cmd += " batch {batch}".format(batch=verb["batch"])
    if verb.get("threshold"):
        cmd += " threshold {threshold}".format(threshold=verb["threshold"])
    if verb.get("immediate"):
        cmd += " {immediate}".format(immediate="immediate")
    if verb.get("protected"):
        cmd += " {protected}".format(protected="protected")
    if verb.get("notify"):
        cmd += " {notify}".format(notify="notify")
    return cmd


class Logging_globalTemplate(NetworkTemplate):
    def __init__(self, lines=None, module=None):
        super(Logging_globalTemplate, self).__init__(lines=lines, tmplt=self, module=module)

    # fmt: off
    PARSERS = [
        {
            "name": "hosts",
            "getval": re.compile(
                r"""
                ^logging\shost
                (\s(?P<hostname>\S+))?
                (\sipv6\s(?P<ipv6>\S+))?
                (\svrf\s(?P<vrf>\S+))?
                (\s(?P<filtered>filtered))?
                (\s(?P<xml>xml))?
                (\sstream\s(?P<stream>\d+))?
                (\ssession-id\s(?P<tag>hostname|ipv4|ipv6))?
                (\ssession-id\sstring\s(?P<text>\S+))?
                (\s(?P<sequence_num_session>sequence-num-session))?
                (\sdiscriminator\s(?P<discriminator>.+$))?
                $""", re.VERBOSE,
            ),
            "setval": tmplt_host,
            "remval": tmplt_host_del,
            "result": {
                "hosts": [{
                    "host": "{{ hostname }}",
                    "ipv6": "{{ ipv6 }}",
                    "discriminator": "{{ discriminator }}",
                    "vrf": "{{ vrf }}",
                    "xml": "{{ True if xml is defined }}",
                    "filtered": "{{ True if filtered is defined }}",
                    "sequence_num_session": "{{ True if sequence_num_session is defined }}",
                    "stream": "{{ stream }}",
                    "session_id": {
                        "tag": "{{ tag }}",
                        "text": "{{ text }}",
                    },
                }],
            },
        },
        {
            "name": "hosts.transport",
            "getval": re.compile(
                r"""
                ^logging\shost
                (\s(?P<hostname>\S+))?
                (\sipv6\s(?P<ipv6>\S+))?
                (\svrf\s(?P<vrf>\S+))?
                (\stransport\s(?P<transport>tcp|udp))?
                (\sport\s(?P<port>\d+))?
                (\s(?P<audit>audit))?
                (\s(?P<xml>xml))?
                (\s(?P<filtered>filtered))?
                (\sdiscriminator\s(?P<discriminator>.+$))?
                (\sstream\s(?P<stream>\d+))?
                (\ssession-id\s(?P<tag>hostname|ipv4|ipv6))?
                (\ssession-id\sstring\s(?P<text>\S+))?
                (\s(?P<sequence_num_session>sequence-num-session))?
                $""", re.VERBOSE,
            ),
            "setval": tmplt_host_transport,
            "remval": tmplt_host_transport_del,
            "result": {
                "hosts": [{
                    "host": "{{ hostname }}",
                    "ipv6": "{{ ipv6 }}",
                    "vrf": "{{ vrf if transport is defined }}",
                    "transport": {
                        "{{ transport }}": {
                            "audit": "{{ True if audit is defined }}",
                            "sequence_num_session": "{{ True if sequence_num_session is defined }}",
                            "xml": "{{ True if xml is defined }}",
                            "discriminator": "{{ discriminator }}",
                            "port": "{{ port }}",
                            "session_id": {
                                "tag": "{{ tag }}",
                                "text": "{{ text }}",
                            },
                            "filtered": "{{ True if filtered is defined }}",
                            "stream": "{{ stream }}",
                        },
                    },
                }],
            },
        },
        {
            "name": "buffered",
            "getval": re.compile(
                r"""
                ^logging\sbuffered
                (\sdiscriminator\s(?P<discriminator>.+))?
                (\s(?P<filtered>filtered))?
                (\s(?P<xml>xml))?
                (\s(?P<size>[1-9][0-9]*))?
                (\s(?P<severity>alerts|critical|debugging|emergencies|errors|informational|notifications|warnings))?
                $""", re.VERBOSE,
            ),
            "setval": tmplt_buffered,
            "result": {
                "buffered": {
                    "filtered": "{{ True if filtered is defined }}",
                    "xml": "{{ True if xml is defined }}",
                    "severity": "{{ severity }}",
                    "size": "{{ size }}",
                    "discriminator": "{{ discriminator }}",
                },
            },
        },
        {
            "name": "buginf",
            "getval": re.compile(
                r"""
                ^logging\s(?P<buginf>buginf)
                $""", re.VERBOSE,
            ),
            "setval": "logging buginf",
            "result": {
                "buginf": "{{ True if buginf is defined }}",
            },
        },
        {
            "name": "cns_events",
            "getval": re.compile(
                r"""
                ^logging\scns-events
                (\s(?P<severity>alerts|critical|debugging|emergencies|errors|informational|notifications|warnings))?
                $""", re.VERBOSE,
            ),
            "setval": "logging cns-events {{ cns_events }}",
            "result": {
                "cns_events": "{{ severity }}",
            },
        },
        {
            "name": "console",
            "getval": re.compile(
                r"""
                ^logging\sconsole
                (\s(?P<filtered>filtered))?
                (\s(?P<xml>xml))?
                (\s(?P<severity>alerts|critical|debugging|emergencies|errors|informational|notifications|warnings|guaranteed$))?
                (\sdiscriminator\s(?P<discriminator>.+$))?
                $""", re.VERBOSE,
            ),
            "setval": tmplt_console,
            "result": {
                "console": {
                    "severity": "{{ severity }}",
                    "discriminator": "{{ discriminator }}",
                    "filtered": "{{ True if filtered is defined }}",
                    "xml": "{{ True if xml is defined }}",
                },
            },
        },
        {
            "name": "count",
            "getval": re.compile(
                r"""
                ^logging\s(?P<count>count)
                $""", re.VERBOSE,
            ),
            "setval": "logging count",
            "result": {
                "count": "{{ True if count is defined }}",
            },
        },
        {
            "name": "delimiter",
            "getval": re.compile(
                r"""
                ^logging\sdelimiter\s(?P<tcp>tcp)
                $""", re.VERBOSE,
            ),
            "setval": "logging delimiter tcp",
            "result": {
                "delimiter": {
                    "tcp": "{{ True if tcp is defined }}",
                },
            },
        },
        {
            "name": "discriminator",
            "getval": re.compile(
                r"""
                ^logging\sdiscriminator\s(?P<discriminator>.+$)
                $""", re.VERBOSE,
            ),
            "setval": "logging discriminator {{ discriminator }}",
            "result": {
                "discriminator": ["{{ discriminator }}"],
            },
        },
        {
            "name": "dmvpn",
            "getval": re.compile(
                r"""
                ^logging\sdmvpn\srate-limit
                (\s(?P<rate>\d+))?
                $""", re.VERBOSE,
            ),
            "setval": "logging dmvpn rate-limit {{ dmvpn.rate_limit }}",
            "result": {
                "dmvpn": {
                    "rate_limit": "{{ rate }}",
                },
            },
        },
        {
            "name": "esm",
            "getval": re.compile(
                r"""
                ^logging\sesm\s(?P<config>config)
                $""", re.VERBOSE,
            ),
            "setval": "logging esm config",
            "result": {
                "esm": {
                    "config": "{{ True if config is defined }}",
                },
            },
        },
        {
            "name": "exception",
            "getval": re.compile(
                r"""
                ^logging\sexception
                \s(?P<exception>[1-9][0-9]*)
                $""", re.VERBOSE,
            ),
            "setval": "logging exception {{ exception }}",
            "result": {
                "exception": "{{ exception }}",
            },
        },
        {
            "name": "facility",
            "getval": re.compile(
                r"""
                ^logging\sfacility
                \s(?P<facility>auth|cron|daemon|kern|local0|local1|local2|local3|local4|local5|local6|local7|lpr|mail|news|sys10|sys11|sys12|sys13|sys14|sys9|syslog|user|uucp)
                $""", re.VERBOSE,
            ),
            "setval": "logging facility {{ facility }}",
            "result": {
                "facility": "{{ facility }}",
            },
        },
        {
            "name": "filter",
            "getval": re.compile(
                r"""
                ^logging\sfilter
                (\s(?P<url>\S+))?
                (\s(?P<order>\d+))?
                (\sargs\s(?P<args>.+$))?
                $""", re.VERBOSE,
            ),
            "setval": tmplt_filter,
            "result": {
                "filter": [{
                    "url": "{{ url }}",
                    "order": "{{ order }}",
                    "args": "{{ args }}",
                }],
            },
        },
        {
            "name": "history.size",
            "getval": re.compile(
                r"""
                ^logging\shistory\ssize
                (\s(?P<size>\d+))
                $""", re.VERBOSE,
            ),
            "setval": "logging history size {{ history.size }}",
            "result": {
                "history": {
                    "size": "{{ size }}",
                },
            },
        },
        {
            "name": "history.severity",
            "getval": re.compile(
                r"""
                ^logging\shistory
                (\s(?P<severity>alerts|critical|debugging|emergencies|errors|informational|notifications|warnings))
                $""", re.VERBOSE,
            ),
            "setval": "logging history {{ history.severity }}",
            "result": {
                "history": {
                    "severity": "{{ severity }}",
                },
            },
        },
        {
            "name": "message_counter",
            "getval": re.compile(
                r"""
                ^logging\smessage-counter
                \s(?P<counter>log|debug|syslog)
                $""", re.VERBOSE,
            ),
            "setval": tmplt_message_counter,
            "result": {
                "message_counter": ["{{ counter }}"],
            },
        },
        {
            "name": "monitor",
            "getval": re.compile(
                r"""
                ^logging\smonitor
                (\s(?P<filtered>filtered))?
                (\s(?P<xml>xml))?
                (\s(?P<severity>alerts|critical|debugging|emergencies|errors|informational|notifications|warnings))?
                (\sdiscriminator\s(?P<discriminator>.+$))?
                $""", re.VERBOSE,
            ),
            "setval": tmplt_monitor,
            "result": {
                "monitor": {
                    "filtered": "{{ True if filtered is defined }}",
                    "xml": "{{ True if xml is defined }}",
                    "severity": "{{ severity }}",
                    "discriminator": "{{ discriminator }}",
                },
            },
        },
        {
            "name": "logging_on",
            "getval": re.compile(
                r"""
                ((?P<negate>no))?
                ((\s|^))?logging
                (\s(?P<on>on))?
                $""", re.VERBOSE,
            ),
            "setval": tmplt_logging_on,
            "remval": "logging on",
            "result": {
                "logging_on": "{{ 'disable' if negate is defined else 'enable' }}",
            },
        },
        {
            "name": "origin_id",
            "getval": re.compile(
                r"""
                ^logging\sorigin-id
                (\s(?P<tag>hostname|ip|ipv6))?
                (\sstring\s(?P<text>.+$))?
                $""", re.VERBOSE,
            ),
            "setval": tmplt_origin_id,
            "result": {
                "origin_id": {
                    "tag": "{{ tag }}",
                    "text": "{{ text }}",
                },
            },
        },
        {
            "name": "persistent",
            "getval": re.compile(
                r"""
                ^logging\spersistent
                (\surl\s(?P<url>\S+))?
                (\ssize\s(?P<size>[1-9][0-9]*))?
                (\sfilesize\s(?P<filesize>[1-9][0-9]*))?
                (\sbatch\s(?P<batch>[1-9][0-9]*))?
                (\sthreshold\s(?P<threshold>[1-9][0-9]*))?
                (\s(?P<immediate>immediate))?
                (\s(?P<protected>protected))?
                (\s(?P<notify>notify))?
                $""", re.VERBOSE,
            ),
            "setval": tmplt_persistent,
            "result": {
                "persistent": {
                    "batch": "{{ batch }}",
                    "filesize": "{{ filesize }}",
                    "immediate": "{{ True if immediate is defined }}",
                    "notify": "{{ True if notify is defined }}",
                    "protected": "{{ True if protected is defined }}",
                    "size": "{{ size }}",
                    "threshold": "{{ threshold }}",
                    "url": "{{ url }}",
                },
            },
        },
        {
            "name": "policy_firewall",
            "getval": re.compile(
                r"""
                ^logging\spolicy-firewall
                (\srate-limit\s(?P<rate>[1-9][0-9]*))?
                $""", re.VERBOSE,
            ),
            "setval": "logging policy-firewall rate-limit {{ policy_firewall.rate_limit }}",
            "result": {
                "policy_firewall": {
                    "rate_limit": "{{ rate }}",
                },
            },
        },
        {
            "name": "queue_limit",
            "getval": re.compile(
                r"""
                ^logging\squeue-limit
                (\s(?P<size>[1-9][0-9]*))?
                (\sesm\s(?P<esm>[1-9][0-9]*))?
                (\strap\s(?P<trap>[1-9][0-9]*))?
                $""", re.VERBOSE,
            ),
            "setval": tmplt_queue_limit,
            "result": {
                "queue_limit": {
                    "size": "{{ size }}",
                    "esm": "{{ esm }}",
                    "trap": "{{ trap }}",
                },
            },
        },
        {
            "name": "rate_limit",
            "getval": re.compile(
                r"""
                ^logging\srate-limit
                (\s(?P<option>all|console))?
                (\s(?P<size>[1-9][0-9]*))?
                (\sexcept\s(?P<except_severity>alerts|critical|debugging|emergencies|errors|informational|notifications|warnings))?
                $""", re.VERBOSE,
            ),
            "setval": tmplt_rate_limit,
            "result": {
                "rate_limit": {
                    "size": "{{ size }}",
                    "all": "{{ True if option == 'all' }}",
                    "console": "{{ True if option == 'console' }}",
                    "except_severity": "{{ except_severity }}",
                },
            },
        },
        {
            "name": "reload",
            "getval": re.compile(
                r"""
                ^logging
                (\s(?P<reload>reload))?
                (\smessage-limit\s(?P<message_limit>[1-9][0-9]*))?
                (\s(?P<severity>alerts|critical|debugging|emergencies|errors|informational|notifications|warnings))?
                $""", re.VERBOSE,
            ),
            "setval": tmplt_reload,
            "result": {
                "reload": {
                    "severity": "{{ severity }}",
                    "message_limit": "{{ message_limit }}",
                },
            },
        },
        {
            "name": "server_arp",
            "getval": re.compile(
                r"""
                ^logging
                \s(?P<server_arp>server-arp)
                $""", re.VERBOSE,
            ),
            "setval": "logging server-arp",
            "result": {
                "server_arp": "{{ True if server_arp is defined }}",
            },
        },
        {
            "name": "snmp_trap",
            "getval": re.compile(
                r"""
                ^logging\ssnmp-trap
                (\s(?P<severity>alerts|critical|debugging|emergencies|errors|informational|notifications|warnings))?
                $""", re.VERBOSE,
            ),
            "setval": "logging snmp-trap {{ snmp_trap }}",
            "result": {
                "snmp_trap": ["{{ severity }}"],
            },
        },
        {
            "name": "source_interface",
            "getval": re.compile(
                r"""
                ^logging\ssource-interface
                (\s(?P<interface>\S+))?
                (\svrf\s(?P<vrf>\S+))?
                $""", re.VERBOSE,
            ),
            "setval": tmplt_source_interface,
            "result": {
                "source_interface": [{
                    "interface": "{{ interface }}",
                    "vrf": "{{ vrf }}",
                }],
            },
        },
        {
            "name": "trap",
            "getval": re.compile(
                r"""
                ^logging\strap
                \s(?P<severity>alerts|critical|debugging|emergencies|errors|informational|notifications|warnings)
                $""", re.VERBOSE,
            ),
            "setval": "logging trap {{ trap }}",
            "result": {
                "trap": "{{ severity }}",
            },
        },
        {
            "name": "trap_include",
            "getval": re.compile(
                r"""
                ^Trap\slogging:\slevel\s(?P<severity>alerts|critical|debugging|emergencies|errors|informational|notifications|warnings),\s*
                (?P<message_lines>\d+)\s+message\slines\slogged$""", re.VERBOSE,
            ),
            "setval": "logging trap {{ trap }}",
            "result": {
                "trap": "{{ severity }}",
            },
        },
        {
            "name": "userinfo",
            "getval": re.compile(
                r"""
                ^logging\s(?P<userinfo>userinfo)
                $""", re.VERBOSE,
            ),
            "setval": "logging userinfo",
            "result": {
                "userinfo": "{{ True if userinfo is defined }}",
            },
        },
    ]
    # fmt: on



#!/usr/bin/python
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_eventgrid_subscription
version_added: "3.13.0"
short_description: Manage Eventgrid Subscriptions
description:
    - Create, update ,delete and replace Eventgrid Subscriptions.
options:
    name:
        description:
            - Name of the subscription.
        required: true
        type: str
    source_resource_id:
        description:
            - Name of a resource id where the subscription exists or will be created.
        required: true
        type: str
    state:
        description:
            - State of the Eventgrid Event Subscription. Use C(present) to create
              or update and C(absent) to delete.
        default: present
        type: str
        choices:
            - absent
            - present


extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_eventgrid_subscription

author:
    - Bill Peck (@p3ck)
'''

EXAMPLES = '''
- name: Subscribe to custom topic
  azure.azcollection.azure_rm_eventgrid_subscription:
    name: subscription-xxxxx20070
    source_resource_id: /subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/MyResourceGroup/providers/Microsoft.EventGrid/topics/topic-xxxxx20070
    endpoint: https://event-grid-site-xxxxx20070.azurewebsites.net/api/updates
'''

RETURN = '''
changed:
    description:
        - Whether the resource is changed.
    returned: always
    type: bool
    sample: true
eventgrid_subscriptions:
    description:
        - List of Eventgrid subscriptions.
    returned: always
    type: complex
    contains:
        destination:
            description:
                - Details of the destination
            returned: always
            type: complex
            contains:
                endpoint_base_url:
                    description:
                        - The base URL of the endpoint
                    returned: always
                    type: str
                    sample: https://event-grid-site-xxxxx20070.azurewebsites.net/api/updates
                endpoint_type:
                    description:
                        - The type of the endpoint
                    returned: always
                    type: str
                    sample: WebHook
                max_events_per_batch:
                    description:
                        - How many events will be batched
                    returned: always
                    type: int
                    sample: 1
                preferred_batch_size_in_kilobytes:
                    description:
                        - The preferred batch size in kilobytes
                    returned: always
                    type: int
                    sample: 64
        event_delivery_schema:
            description:
                - The type of Eventgrid Schema
            returned: always
            type: str
            sample: EventGridSchema
        filter:
            description:
                - Filter settings for which events to publish to endpoint
            returned: always
            type: complex
            contains:
                subject_begins_with:
                    description:
                        - Filter on Subject beginning with this value
                    returned: always
                    type: str
                subject_ends_with:
                    description:
                        - Filter on Subject ending with this value
                    returned: always
                    type: str
        id:
            description:
                - The ID of the Eventgrid Subscription
            returned: always
            type: str
            sample: /subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/MyResourceGroup/providers/Microsoft.EventGrid/topics/topic-xxxxx20070/providers/Microsoft.EventGrid/eventSubscriptions/subscription-xxxxx20070 # nopep8: E501
        name:
            description:
                - The name of the Eventgrid Subscription
            returned: always
            type: str
            sample: subscription-xxxxx20070
        provisioning_state:
            description:
                - The provisioning state of this Eventgrid Subscription
            returned: always
            type: str
            sample: Succeeded
        retry_policy:
            description:
                - Retry policy for this Eventgrid Subscription
            returned: always
            type: complex
            contains:
                event_time_to_live_in_minutes:
                    description:
                        - Event time to live in minutes
                    returned: always
                    type: int
                    sample: 1440
                max_delivery_attempts:
                    description:
                        - Max attempts to deliver to endpoint
                    returned: always
                    type: int
                    sample: 30
        topic:
            description:
                - The source resource id that the Eventgrid Subscription is subscribed to
            returned: always
            type: str
            sample: /subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/MyResourceGroup/providers/microsoft.eventgrid/topics/topic-xxxxx20070
        type:
            description:
                - The type of Eventgrid Subscription
            returned: always
            type: str
            sample: Microsoft.EventGrid/eventSubscriptions
'''


from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ext import AzureRMModuleBaseExt
from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_eventgrid import (
    eventgrid_subscription_module_arg_spec, get_event_subscription_info,
    update_event_subscription_internal, ConfigError)
try:
    from azure.mgmt.eventgrid import EventGridManagementClient
    from azure.core.polling import LROPoller
    from azure.core.exceptions import ResourceNotFoundError
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMEventgridSubscription(AzureRMModuleBaseExt):

    @property
    def client(self):
        self.log('Getting client')
        if not self._client:
            self._client = self.get_mgmt_svc_client(EventGridManagementClient,
                                                    base_url=self._cloud_environment.endpoints.resource_manager)
        return self._client

    def __init__(self):

        self.module_arg_spec = dict(
            name=dict(
                type='str',
                required=True
            ),
            source_resource_id=dict(
                type='str',
                required=True,
            ),
            state=dict(
                type='str',
                default='present',
                choices=['present', 'absent']
            ),
        )
        self.module_arg_spec.update(eventgrid_subscription_module_arg_spec)

        self._client = None

        self.name = None
        self.source_resource_id = None
        self.endpoint = None
        self.endpoint_type = None
        self.included_event_types = None
        self.subject_begins_with = None
        self.subject_ends_with = None
        self.is_subject_case_sensitive = None
        self.max_delivery_attempts = None
        self.event_ttl = None
        self.max_events_per_batch = None
        self.preferred_batch_size_in_kilobytes = None
        self.event_delivery_schema = None
        self.deadletter_endpoint = None
        self.labels = None
        self.expiration_date = None
        self.advanced_filter = None
        self.azure_active_directory_tenant_id = None
        self.azure_active_directory_application_id_or_uri = None
        self.storage_queue_msg_ttl = None
        self.enable_advanced_filtering_on_arrays = None
        self.delivery_attribute_mapping = None

        self.state = None

        self.results = dict(
            changed=False,
            eventgrid_event_subscription=dict()
        )

        required_together = None
        required_if = None
#        required_together = [['extended_location_name', 'extended_location_type']]
#        required_if = [
#            ("kind", "AzureArc", ["extended_location_name"]),
#        ]

        super(AzureRMEventgridSubscription, self).__init__(self.module_arg_spec,
                                                           supports_tags=False,
                                                           supports_check_mode=True,
                                                           required_together=required_together,
                                                           required_if=required_if)

    def exec_module(self, **kwargs):

        for key in list(self.module_arg_spec.keys()):
            setattr(self, key, kwargs[key])

        changed = False

        if self.state == 'present':
            event_subscription_info = self.get_event_subscription(self.source_resource_id,
                                                                  self.name)
            if event_subscription_info:
                try:
                    params = update_event_subscription_internal(
                        instance=event_subscription_info,
                        endpoint=self.endpoint,
                        endpoint_type=self.update_endpoint_type,
                        subject_begins_with=self.subject_begins_with,
                        subject_ends_with=self.subject_ends_with,
                        included_event_types=self.included_event_types,
                        advanced_filter=self.advanced_filter,
                        labels=self.labels,
                        deadletter_endpoint=self.deadletter_endpoint,
                        delivery_identity=None,
                        delivery_identity_endpoint=None,
                        delivery_identity_endpoint_type=None,
                        deadletter_identity=None,
                        deadletter_identity_endpoint=None,
                        storage_queue_msg_ttl=self.storage_queue_msg_ttl,
                        enable_advanced_filtering_on_arrays=self.enable_advanced_filtering_on_arrays,
                        delivery_attribute_mapping=self.delivery_attribute_mapping)
                except ConfigError as exc:
                    self.fail(str(exc))

                self.results['compare'] = []
                changed = not self.default_compare({},
                                                   self.format_event_subscription(params),
                                                   self.format_event_subscription(event_subscription_info),
                                                   '',
                                                   self.results)
                if changed and not self.check_mode:
                    event_subscription_info = self.begin_update_event_subscription(self.source_resource_id,
                                                                                   self.name,
                                                                                   params)
                    response = self.format_event_subscription(event_subscription_info)

            else:
                changed = True

                try:
                    event_subscription_info = get_event_subscription_info(
                        endpoint=self.endpoint,
                        endpoint_type=self.endpoint_type,
                        included_event_types=self.included_event_types,
                        subject_begins_with=self.subject_begins_with,
                        subject_ends_with=self.subject_ends_with,
                        is_subject_case_sensitive=self.is_subject_case_sensitive,
                        max_delivery_attempts=self.max_delivery_attempts,
                        event_ttl=self.event_ttl,
                        max_events_per_batch=self.max_events_per_batch,
                        preferred_batch_size_in_kilobytes=self.preferred_batch_size_in_kilobytes,
                        event_delivery_schema=self.event_delivery_schema,
                        deadletter_endpoint=self.deadletter_endpoint,
                        labels=self.labels,
                        expiration_date=self.expiration_date,
                        advanced_filter=self.advanced_filter,
                        azure_active_directory_tenant_id=self.azure_active_directory_tenant_id,
                        azure_active_directory_application_id_or_uri=self.azure_active_directory_application_id_or_uri,
                        delivery_identity=None,
                        delivery_identity_endpoint=None,
                        delivery_identity_endpoint_type=None,
                        deadletter_identity=None,
                        deadletter_identity_endpoint=None,
                        storage_queue_msg_ttl=self.storage_queue_msg_ttl,
                        enable_advanced_filtering_on_arrays=self.enable_advanced_filtering_on_arrays,
                        delivery_attribute_mapping=self.delivery_attribute_mapping)
                except ConfigError as exc:
                    self.fail(str(exc))

                if not self.check_mode:
                    event_subscription_info = self.begin_create_or_update_event_subscription(self.source_resource_id,
                                                                                             self.name,
                                                                                             event_subscription_info)
                response = self.format_event_subscription(event_subscription_info)
        else:
            event_subscription_info = self.get_event_subscription(self.source_resource_id,
                                                                  self.name)
            if event_subscription_info:
                changed = True
                if not self.check_mode:
                    response = self.delete_event_subscription(self.source_resource_id,
                                                              self.name)
                else:
                    response = self.format_event_subscription(event_subscription_info)

        self.results['changed'] = changed
        self.results['eventgrid_event_subscription'] = response

        return self.results

    def begin_create_or_update_event_subscription(self,
                                                  source_resource_id,
                                                  event_subscription_name,
                                                  event_subscription_info):
        self.log('Creates or updates the Eventgrid Event Subscription.')

        try:
            response = self.client.event_subscriptions.begin_create_or_update(source_resource_id,
                                                                              event_subscription_name,
                                                                              event_subscription_info)
            if isinstance(response, LROPoller):
                response = self.get_poller_result(response)
        except Exception as exc:
            self.fail('Creates or updates the Eventgrid Event Subscription got Exception as as {0}'.format(str(exc)))
        return response

    def begin_update_event_subscription(self,
                                        source_resource_id,
                                        event_subscription_name,
                                        event_subscription_update_parameters):
        self.log('Selectively updates the Event Subscription.')

        try:
            response = self.client.event_subscriptions.begin_update(source_resource_id,
                                                                    event_subscription_name,
                                                                    event_subscription_update_parameters)
            if isinstance(response, LROPoller):
                response = self.get_poller_result(response)
        except Exception as exc:
            self.fail('Selectively update the Eventgrid Event Subscription got Excption as {0}'.format(str(exc)))
        return response

    def delete_event_subscription(self, source_resource_id, event_subsription_name):
        self.log('Deletes the entire Eventgrid Event Subscription.')
        try:
            response = self.client.event_subscriptions.begin_delete(source_resource_id, event_subsription_name)
            if isinstance(response, LROPoller):
                response = self.get_poller_result(response)
        except Exception as exc:
            self.fail('Delete Event Subscription got Excetion as {0}'.format(str(exc)))

    def get_event_subscription(self, source_resource_id, event_subscription_name):
        self.log('Get properties for {0}'.format(source_resource_id))
        try:
            return self.client.event_subscriptions.get(source_resource_id, event_subscription_name)
        except ResourceNotFoundError:
            self.log('Did not find the event_subscription.')
            return False
        except Exception as exc:
            self.fail('Error trying to get event_subscription got Excetion as {0}'.format(str(exc)))

    def format_event_subscription(self, event_subscription_data):
        results = event_subscription_data and event_subscription_data.as_dict() or {}
        return results


def main():
    AzureRMEventgridSubscription()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Ansible module to manage CheckPoint Firewall (c) 2019
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = """
---
module: cp_mgmt_resource_uri
short_description: Manages resource-uri objects on Checkpoint over Web Services API
description:
  - Manages resource-uri objects on Checkpoint devices including creating, updating and removing objects.
  - All operations are performed over Web Services API.
  - Available from R82 management version.
version_added: "6.0.0"
author: "Eden Brillant (@chkp-edenbr)"
options:
  name:
    description:
      - Object name.
    type: str
    required: True
  use_this_resource_to:
    description:
      - Select the use of the URI resource.
    type: str
    choices: ['enforce_uri_capabilities', 'optimize_url_logging', 'enhance_ufp_performance']
  connection_methods:
    description:
      - Connection methods.
    type: dict
    suboptions:
      transparent:
        description:
          - The security server is invisible to the client that originates the connection, and to the server. The Transparent connection method is
            the most secure.
        type: bool
      proxy:
        description:
          - The Resource is applied when people specify the Check Point Security Gateway as a proxy in their browser.
        type: bool
      tunneling:
        description:
          - The Resource is applied when people specify the Security Gateway as a proxy in their browser, and is used for connections where
            Security Gateway cannot examine the contents of the packets, not even the URL.
        type: bool
  uri_match_specification_type:
    description:
      - The type can be Wild Cards or UFP, where a UFP server holds categories of forbidden web sites.
    type: str
    choices: ['wildcards', 'ufp']
  exception_track:
    description:
      - Configures how to track connections that match this rule but fail the content security checks. An example of an exception is a connection with
        an unsupported scheme or method.
    type: str
    choices: ['none', 'exception log', 'exception alert']
  match_ufp:
    description:
      - Match-Ufp settings.
    type: dict
    suboptions:
      server:
        description:
          - The UID or Name of the UFP server that is an OPSEC certified third party application that checks URLs against a list of permitted categories.
        type: str
      caching_control:
        description:
          - Specifies if and how caching is to be enabled.
        type: str
        choices: ['security_gateway_one_request', 'security_gateway_two_requests', 'no_caching', 'ufp_server']
      ignore_ufp_server_after_failure:
        description:
          - The UFP server will be ignored after numerous UFP server connections were unsuccessful.
        type: bool
      number_of_failures_before_ignore:
        description:
          - Signifies at what point the UFP server should be ignored.
        type: int
      timeout_before_reconnecting:
        description:
          - The amount of time that must pass before a UFP server connection should be attempted.
        type: int
  match_wildcards:
    description:
      - Match-Wildcards settings.
    type: dict
    suboptions:
      schemes:
        description:
          - Select the URI Schemes to which this resource applies.
        type: dict
        suboptions:
          http:
            description:
              - Http scheme.
            type: bool
          ftp:
            description:
              - Ftp scheme.
            type: bool
          gopher:
            description:
              - Gopher scheme.
            type: bool
          mailto:
            description:
              - Mailto scheme.
            type: bool
          news:
            description:
              - News scheme.
            type: bool
          wais:
            description:
              - Wais scheme.
            type: bool
          other:
            description:
              - You can specify another scheme in the Other field. You can use wildcards.
            type: str
      methods:
        description:
          - Select the URI Schemes to which this resource applies.
        type: dict
        suboptions:
          get:
            description:
              - GET method.
            type: bool
          post:
            description:
              - POST method.
            type: bool
          head:
            description:
              - HEAD method.
            type: bool
          put:
            description:
              - PUT method.
            type: bool
          other:
            description:
              - You can specify another method in the Other field. You can use wildcards.
            type: str
      host:
        description:
          - The functionality of the Host parameter depends on the DNS setup of the addressed server. For the host, only the IP address or the
            full DNS name should be used.
        type: str
      path:
        description:
          - Name matching is based on appending the file name in the request to the current working directory (unless the file name is already a
            full path name) and comparing the result to the path specified in the Resource definition.
        type: str
      query:
        description:
          - The parameters that are sent to the URI when it is accessed.
        type: str
  action:
    description:
      - Action settings.
    type: dict
    suboptions:
      replacement_uri:
        description:
          - If the Action in a rule which uses this resource is Drop or Reject, then the Replacement URI is displayed instead of the one requested by the user.
        type: str
      strip_script_tags:
        description:
          - Strip JAVA scripts.
        type: bool
      strip_applet_tags:
        description:
          - Strip JAVA applets.
        type: bool
      strip_activex_tags:
        description:
          - Strip activeX tags.
        type: bool
      strip_ftp_links:
        description:
          - Strip ftp links.
        type: bool
      strip_port_strings:
        description:
          - Strip ports.
        type: bool
  cvp:
    description:
      - CVP settings.
    type: dict
    suboptions:
      enable_cvp:
        description:
          - Select to enable the Content Vectoring Protocol.
        type: bool
      server:
        description:
          - The UID or Name of the CVP server, make sure the CVP server is already be defined as an OPSEC Application.
        type: str
      allowed_to_modify_content:
        description:
          - Configures the CVP server to inspect but not modify content.
        type: bool
      send_http_headers_to_cvp:
        description:
          - Select, if you would like the CVP server to check the HTTP headers of the message packets.
        type: bool
      reply_order:
        description:
          - Designates when the CVP server returns data to the Security Gateway security server.
        type: str
        choices: ['return_data_after_content_is_approved', 'return_data_before_content_is_approved']
      send_http_request_to_cvp:
        description:
          - Used to protect against undesirable content in the HTTP request, for example, when inspecting peer-to-peer connections.
        type: bool
      send_only_unsafe_file_types:
        description:
          - Improves the performance of the CVP server. This option does not send to the CVP server traffic that is considered safe.
        type: bool
  soap:
    description:
      - SOAP settings.
    type: dict
    suboptions:
      inspection:
        description:
          - Allow all SOAP Requests, or Allow only SOAP requests specified in the following file-id.
        type: str
        choices: ['allow_all_soap_requests', 'allow_soap_requests_as_specified_in_file']
      file_id:
        description:
          - A file containing SOAP requests.
        type: str
        choices: ['scheme1', 'scheme2', 'scheme3', 'scheme4', 'scheme5', 'scheme6', 'scheme7', 'scheme8', 'scheme9', 'scheme10']
      track_connections:
        description:
          - The method of tracking SOAP connections.
        type: str
        choices: ['none', 'log', 'popup_alert', 'mail_alert', 'snmp_trap_alert', 'user_defined_alert_no', 'user_defined_alert_no', 'user_defined_alert_no']
  tags:
    description:
      - Collection of tag identifiers.
    type: list
    elements: str
  color:
    description:
      - Color of the object. Should be one of existing colors.
    type: str
    choices: ['aquamarine', 'black', 'blue', 'crete blue', 'burlywood', 'cyan', 'dark green', 'khaki', 'orchid', 'dark orange', 'dark sea green',
             'pink', 'turquoise', 'dark blue', 'firebrick', 'brown', 'forest green', 'gold', 'dark gold', 'gray', 'dark gray', 'light green', 'lemon chiffon',
             'coral', 'sea green', 'sky blue', 'magenta', 'purple', 'slate blue', 'violet red', 'navy blue', 'olive', 'orange', 'red', 'sienna', 'yellow']
  comments:
    description:
      - Comments string.
    type: str
  details_level:
    description:
      - The level of detail for some of the fields in the response can vary from showing only the UID value of the object to a fully detailed
        representation of the object.
    type: str
    choices: ['uid', 'standard', 'full']
  ignore_warnings:
    description:
      - Apply changes ignoring warnings.
    type: bool
  ignore_errors:
    description:
      - Apply changes ignoring errors. You won't be able to publish such a changes. If ignore-warnings flag was omitted - warnings will also be ignored.
    type: bool
extends_documentation_fragment: check_point.mgmt.checkpoint_objects
"""

EXAMPLES = """
- name: add-resource-uri
  cp_mgmt_resource_uri:
    connection_methods:
      transparent: 'false'
      tunneling: 'true'
    match_wildcards:
      host: hostName
      path: pathName
    name: newUriResource
    state: present
    uri_match_specification_type: wildcards
    use_this_resource_to: optimize_url_logging

- name: set-resource-uri
  cp_mgmt_resource_uri:
    connection_methods:
      transparent: 'false'
      tunneling: 'true'
    match_wildcards:
      host: hostName
      path: pathName
    name: newUriResource
    state: present
    uri_match_specification_type: wildcards
    use_this_resource_to: optimize_url_logging

- name: delete-resource-uri
  cp_mgmt_resource_uri:
    name: newUriResource
    state: absent
"""

RETURN = """
cp_mgmt_resource_uri:
  description: The checkpoint object created or updated.
  returned: always, except when deleting the object.
  type: dict
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.check_point.mgmt.plugins.module_utils.checkpoint import checkpoint_argument_spec_for_objects, api_call


def main():
    argument_spec = dict(
        name=dict(type='str', required=True),
        use_this_resource_to=dict(type='str', choices=['enforce_uri_capabilities', 'optimize_url_logging', 'enhance_ufp_performance']),
        connection_methods=dict(type='dict', options=dict(
            transparent=dict(type='bool'),
            proxy=dict(type='bool'),
            tunneling=dict(type='bool')
        )),
        uri_match_specification_type=dict(type='str', choices=['wildcards', 'ufp']),
        exception_track=dict(type='str', choices=['none', 'exception log', 'exception alert']),
        match_ufp=dict(type='dict', options=dict(
            server=dict(type='str'),
            caching_control=dict(type='str', choices=['security_gateway_one_request', 'security_gateway_two_requests', 'no_caching', 'ufp_server']),
            ignore_ufp_server_after_failure=dict(type='bool'),
            number_of_failures_before_ignore=dict(type='int'),
            timeout_before_reconnecting=dict(type='int')
        )),
        match_wildcards=dict(type='dict', options=dict(
            schemes=dict(type='dict', options=dict(
                http=dict(type='bool'),
                ftp=dict(type='bool'),
                gopher=dict(type='bool'),
                mailto=dict(type='bool'),
                news=dict(type='bool'),
                wais=dict(type='bool'),
                other=dict(type='str')
            )),
            methods=dict(type='dict', options=dict(
                get=dict(type='bool'),
                post=dict(type='bool'),
                head=dict(type='bool'),
                put=dict(type='bool'),
                other=dict(type='str')
            )),
            host=dict(type='str'),
            path=dict(type='str'),
            query=dict(type='str')
        )),
        action=dict(type='dict', options=dict(
            replacement_uri=dict(type='str'),
            strip_script_tags=dict(type='bool'),
            strip_applet_tags=dict(type='bool'),
            strip_activex_tags=dict(type='bool'),
            strip_ftp_links=dict(type='bool'),
            strip_port_strings=dict(type='bool')
        )),
        cvp=dict(type='dict', options=dict(
            enable_cvp=dict(type='bool'),
            server=dict(type='str'),
            allowed_to_modify_content=dict(type='bool'),
            send_http_headers_to_cvp=dict(type='bool'),
            reply_order=dict(type='str', choices=['return_data_after_content_is_approved', 'return_data_before_content_is_approved']),
            send_http_request_to_cvp=dict(type='bool'),
            send_only_unsafe_file_types=dict(type='bool')
        )),
        soap=dict(type='dict', options=dict(
            inspection=dict(type='str', choices=['allow_all_soap_requests', 'allow_soap_requests_as_specified_in_file']),
            file_id=dict(type='str', choices=['scheme1', 'scheme2', 'scheme3', 'scheme4', 'scheme5', 'scheme6', 'scheme7', 'scheme8', 'scheme9', 'scheme10']),
            track_connections=dict(type='str', choices=['none', 'log', 'popup_alert', 'mail_alert',
                                                        'snmp_trap_alert', 'user_defined_alert_no', 'user_defined_alert_no', 'user_defined_alert_no'])
        )),
        tags=dict(type='list', elements='str'),
        color=dict(type='str', choices=['aquamarine', 'black', 'blue', 'crete blue', 'burlywood', 'cyan', 'dark green',
                                        'khaki', 'orchid', 'dark orange', 'dark sea green', 'pink', 'turquoise', 'dark blue', 'firebrick', 'brown',
                                        'forest green', 'gold', 'dark gold', 'gray', 'dark gray', 'light green', 'lemon chiffon', 'coral', 'sea green',
                                        'sky blue', 'magenta', 'purple', 'slate blue', 'violet red', 'navy blue', 'olive', 'orange', 'red', 'sienna',
                                        'yellow']),
        comments=dict(type='str'),
        details_level=dict(type='str', choices=['uid', 'standard', 'full']),
        ignore_warnings=dict(type='bool'),
        ignore_errors=dict(type='bool')
    )
    argument_spec.update(checkpoint_argument_spec_for_objects)

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=True)
    api_call_object = 'resource-uri'

    result = api_call(module, api_call_object)
    module.exit_json(**result)


if __name__ == '__main__':
    main()



#!/usr/bin/python

# Copyright (c) 2020, Florent Madiot (scodeman@scode.io)
# Based on code:
# Copyright (c) 2019, Markus Bergholz (markuman@gmail.com)
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
module: gitlab_group_variable
short_description: Creates, updates, or deletes GitLab groups variables
version_added: 1.2.0
description:
  - Creates a group variable if it does not exist.
  - When a group variable does exist and is not hidden, its value is updated when the values are different.
    When a group variable does exist and is hidden, its value is updated. In this case, the module is B(not idempotent).
  - Variables which are untouched in the playbook, but are not untouched in the GitLab group, they stay untouched (O(purge=false))
    or are deleted (O(purge=true)).
author:
  - Florent Madiot (@scodeman)
requirements:
  - python-gitlab python module
extends_documentation_fragment:
  - community.general._auth_basic
  - community.general._gitlab
  - community.general._attributes

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none

options:
  state:
    description:
      - Create or delete group variable.
    default: present
    type: str
    choices: ["present", "absent"]
  group:
    description:
      - The path and name of the group.
    required: true
    type: str
  purge:
    description:
      - When set to V(true), delete all variables which are not untouched in the task.
    default: false
    type: bool
  vars:
    description:
      - When the list element is a simple key-value pair, C(masked), C(hidden), C(raw), and C(protected) are set to V(false).
      - When the list element is a dict with the keys C(value), C(masked), C(hidden), C(raw), and C(protected), the user can have full
        control about whether a value should be masked, hidden, raw, protected, or a combination.
      - Support for group variables requires GitLab >= 9.5.
      - Support for environment_scope requires GitLab Premium >= 13.11.
      - Support for protected values requires GitLab >= 9.3.
      - Support for masked values requires GitLab >= 11.10.
      - Support for hidden values requires GitLab >= 17.4, and was added in community.general 11.3.0.
      - Support for raw values requires GitLab >= 15.7.
      - A C(value) must be a string or a number.
      - Field C(variable_type) must be a string with either V(env_var), which is the default, or V(file).
      - When a value is masked, it must be in Base64 and have a length of at least 8 characters. See GitLab documentation
        on acceptable values for a masked variable (U(https://docs.gitlab.com/ce/ci/variables/#masked-variables)).
    default: {}
    type: dict
  variables:
    version_added: 4.5.0
    description:
      - A list of dictionaries that represents CI/CD variables.
      - This modules works internal with this structure, even if the older O(vars) parameter is used.
    default: []
    type: list
    elements: dict
    suboptions:
      name:
        description:
          - The name of the variable.
        type: str
        required: true
      value:
        description:
          - The variable value.
          - Required when O(state=present).
        type: str
      description:
        description:
          - A description for the variable.
          - Support for descriptions requires GitLab >= 16.2.
        type: str
        version_added: '11.4.0'
      masked:
        description:
          - Whether variable value is masked or not.
        type: bool
        default: false
      hidden:
        description:
          - Whether variable value is hidden or not.
          - Implies C(masked).
          - Support for hidden values requires GitLab >= 17.4.
        type: bool
        default: false
        version_added: '11.3.0'
      protected:
        description:
          - Whether variable value is protected or not.
        type: bool
        default: false
      raw:
        description:
          - Whether variable value is raw or not.
          - Support for raw values requires GitLab >= 15.7.
        type: bool
        default: false
        version_added: '7.4.0'
      variable_type:
        description:
          - Whether a variable is an environment variable (V(env_var)) or a file (V(file)).
        type: str
        choices: ["env_var", "file"]
        default: env_var
      environment_scope:
        description:
          - The scope for the variable.
        type: str
        default: '*'
"""


EXAMPLES = r"""
- name: Set or update some CI/CD variables
  community.general.gitlab_group_variable:
    api_url: https://gitlab.com
    api_token: secret_access_token
    group: scodeman/testgroup/
    purge: false
    variables:
      - name: ACCESS_KEY_ID
        value: abc123
      - name: SECRET_ACCESS_KEY
        value: 3214cbad
        masked: true
        protected: true
        variable_type: env_var
        environment_scope: production

- name: Set or update some CI/CD variables with raw value
  community.general.gitlab_group_variable:
    api_url: https://gitlab.com
    api_token: secret_access_token
    group: scodeman/testgroup/
    purge: false
    vars:
      ACCESS_KEY_ID: abc123
      SECRET_ACCESS_KEY:
        value: 3214cbad
        masked: true
        protected: true
        raw: true
        variable_type: env_var
        environment_scope: '*'

- name: Set or update some CI/CD variables with expandable value
  community.general.gitlab_group_variable:
    api_url: https://gitlab.com
    api_token: secret_access_token
    group: scodeman/testgroup/
    purge: false
    vars:
      ACCESS_KEY_ID: abc123
      SECRET_ACCESS_KEY:
        value: '$MY_OTHER_VARIABLE'
        masked: true
        protected: true
        raw: false
        variable_type: env_var
        environment_scope: '*'

- name: Delete one variable
  community.general.gitlab_group_variable:
    api_url: https://gitlab.com
    api_token: secret_access_token
    group: scodeman/testgroup/
    state: absent
    vars:
      ACCESS_KEY_ID: abc123
"""

RETURN = r"""
group_variable:
  description: Four lists of the variablenames which were added, updated, removed or exist.
  returned: always
  type: dict
  contains:
    added:
      description: A list of variables which were created.
      returned: always
      type: list
      sample: ["ACCESS_KEY_ID", "SECRET_ACCESS_KEY"]
    untouched:
      description: A list of variables which exist.
      returned: always
      type: list
      sample: ["ACCESS_KEY_ID", "SECRET_ACCESS_KEY"]
    removed:
      description: A list of variables which were deleted.
      returned: always
      type: list
      sample: ["ACCESS_KEY_ID", "SECRET_ACCESS_KEY"]
    updated:
      description: A list of variables whose values were changed.
      returned: always
      type: list
      sample: ["ACCESS_KEY_ID", "SECRET_ACCESS_KEY"]
"""

from ansible.module_utils.api import basic_auth_argument_spec
from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._gitlab import (
    auth_argument_spec,
    filter_returned_variables,
    gitlab_authentication,
    list_all_kwargs,
    vars_to_variables,
)


class GitlabGroupVariables:
    def __init__(self, module, gitlab_instance):
        self.repo = gitlab_instance
        self.group = self.get_group(module.params["group"])
        self._module = module

    def get_group(self, group_name):
        return self.repo.groups.get(group_name)

    def list_all_group_variables(self):
        return list(self.group.variables.list(**list_all_kwargs))

    def create_variable(self, var_obj):
        if self._module.check_mode:
            return True
        var = {
            "key": var_obj.get("key"),
            "value": var_obj.get("value"),
            "description": var_obj.get("description"),
            "masked": var_obj.get("masked"),
            "masked_and_hidden": var_obj.get("hidden"),
            "protected": var_obj.get("protected"),
            "raw": var_obj.get("raw"),
            "variable_type": var_obj.get("variable_type"),
        }
        if var_obj.get("environment_scope") is not None:
            var["environment_scope"] = var_obj.get("environment_scope")

        self.group.variables.create(var)
        return True

    def update_variable(self, var_obj):
        if self._module.check_mode:
            return True
        self.delete_variable(var_obj)
        self.create_variable(var_obj)
        return True

    def delete_variable(self, var_obj):
        if self._module.check_mode:
            return True
        self.group.variables.delete(var_obj.get("key"), filter={"environment_scope": var_obj.get("environment_scope")})
        return True


def compare(requested_variables, existing_variables, state):
    # we need to do this, because it was determined in a previous version - more or less buggy
    # basically it is not necessary and might results in more/other bugs!
    # but it is required  and only relevant for check mode!!
    # logic represents state 'present' when not purge. all other can be derived from that
    # untouched => equal in both
    # updated => name and scope are equal
    # added => name and scope does not exist
    untouched = list()
    updated = list()
    added = list()

    if state == "present":
        existing_key_scope_vars = list()
        for item in existing_variables:
            existing_key_scope_vars.append({"key": item.get("key"), "environment_scope": item.get("environment_scope")})

        for var in requested_variables:
            if var in existing_variables:
                untouched.append(var)
            else:
                compare_item = {"key": var.get("name"), "environment_scope": var.get("environment_scope")}
                if compare_item in existing_key_scope_vars:
                    updated.append(var)
                else:
                    added.append(var)

    return untouched, updated, added


def native_python_main(this_gitlab, purge, requested_variables, state, module):
    change = False
    return_value = dict(added=list(), updated=list(), removed=list(), untouched=list())

    gitlab_keys = this_gitlab.list_all_group_variables()
    before = [x.attributes for x in gitlab_keys]

    gitlab_keys = this_gitlab.list_all_group_variables()
    existing_variables = filter_returned_variables(gitlab_keys)

    for item in requested_variables:
        item["key"] = item.pop("name")
        item["value"] = str(item.get("value"))
        if item.get("protected") is None:
            item["protected"] = False
        if item.get("raw") is None:
            item["raw"] = False
        if item.get("masked") is None:
            item["masked"] = False
        if item.get("hidden") is None:
            item["hidden"] = False
        if item.get("environment_scope") is None:
            item["environment_scope"] = "*"
        if item.get("variable_type") is None:
            item["variable_type"] = "env_var"

    if module.check_mode:
        untouched, updated, added = compare(requested_variables, existing_variables, state)

    if state == "present":
        add_or_update = [x for x in requested_variables if x not in existing_variables]
        for item in add_or_update:
            try:
                if this_gitlab.create_variable(item):
                    return_value["added"].append(item)

            except Exception:
                if this_gitlab.update_variable(item):
                    return_value["updated"].append(item)

        if purge:
            # refetch and filter
            gitlab_keys = this_gitlab.list_all_group_variables()
            existing_variables = filter_returned_variables(gitlab_keys)

            remove = [x for x in existing_variables if x not in requested_variables]
            for item in remove:
                if this_gitlab.delete_variable(item):
                    return_value["removed"].append(item)

    elif state == "absent":
        # value, type, and description do not matter on removing variables.
        keys_ignored_on_deletion = ["value", "variable_type", "description"]
        for key in keys_ignored_on_deletion:
            for item in existing_variables:
                item.pop(key)
            for item in requested_variables:
                item.pop(key)

        if not purge:
            remove_requested = [x for x in requested_variables if x in existing_variables]
            for item in remove_requested:
                if this_gitlab.delete_variable(item):
                    return_value["removed"].append(item)

        else:
            for item in existing_variables:
                if this_gitlab.delete_variable(item):
                    return_value["removed"].append(item)

    if module.check_mode:
        return_value = dict(added=added, updated=updated, removed=return_value["removed"], untouched=untouched)

    if len(return_value["added"] + return_value["removed"] + return_value["updated"]) > 0:
        change = True

    gitlab_keys = this_gitlab.list_all_group_variables()
    after = [x.attributes for x in gitlab_keys]

    return change, return_value, before, after


def main():
    argument_spec = basic_auth_argument_spec()
    argument_spec.update(auth_argument_spec())
    argument_spec.update(
        group=dict(type="str", required=True),
        purge=dict(type="bool", default=False),
        vars=dict(type="dict", default=dict(), no_log=True),
        # please mind whenever changing the variables dict to also change module_utils/gitlab.py's
        # KNOWN dict in filter_returned_variables or bad evil will happen
        variables=dict(
            type="list",
            elements="dict",
            default=list(),
            options=dict(
                name=dict(type="str", required=True),
                value=dict(type="str", no_log=True),
                description=dict(type="str"),
                masked=dict(type="bool", default=False),
                hidden=dict(type="bool", default=False),
                protected=dict(type="bool", default=False),
                raw=dict(type="bool", default=False),
                environment_scope=dict(type="str", default="*"),
                variable_type=dict(type="str", default="env_var", choices=["env_var", "file"]),
            ),
        ),
        state=dict(type="str", default="present", choices=["absent", "present"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=[
            ["api_username", "api_token"],
            ["api_username", "api_oauth_token"],
            ["api_username", "api_job_token"],
            ["api_token", "api_oauth_token"],
            ["api_token", "api_job_token"],
            ["vars", "variables"],
        ],
        required_together=[
            ["api_username", "api_password"],
        ],
        required_one_of=[["api_username", "api_token", "api_oauth_token", "api_job_token"]],
        supports_check_mode=True,
    )

    # check prerequisites and connect to gitlab server
    gitlab_instance = gitlab_authentication(module)

    purge = module.params["purge"]
    var_list = module.params["vars"]
    state = module.params["state"]

    if var_list:
        variables = vars_to_variables(var_list, module)
    else:
        variables = module.params["variables"]

    if state == "present":
        if any(x["value"] is None for x in variables):
            module.fail_json(msg="value parameter is required in state present")

    this_gitlab = GitlabGroupVariables(module=module, gitlab_instance=gitlab_instance)

    changed, raw_return_value, before, after = native_python_main(this_gitlab, purge, variables, state, module)

    # postprocessing
    for item in after:
        item.pop("group_id")
        item["name"] = item.pop("key")
    for item in before:
        item.pop("group_id")
        item["name"] = item.pop("key")

    untouched_key_name = "key"
    if not module.check_mode:
        untouched_key_name = "name"
        raw_return_value["untouched"] = [x for x in before if x in after]

    added = [x.get("key") for x in raw_return_value["added"]]
    updated = [x.get("key") for x in raw_return_value["updated"]]
    removed = [x.get("key") for x in raw_return_value["removed"]]
    untouched = [x.get(untouched_key_name) for x in raw_return_value["untouched"]]
    return_value = dict(added=added, updated=updated, removed=removed, untouched=untouched)

    module.exit_json(changed=changed, group_variable=return_value)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2018, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: cloudformation_stack_set
version_added: 1.0.0
short_description: Manage groups of CloudFormation stacks
description:
  - Launches/updates/deletes AWS CloudFormation Stack Sets.
notes:
  - To make an individual stack, you want the M(amazon.aws.cloudformation) module.
options:
  name:
    description:
      - Name of the CloudFormation stack set.
    required: true
    type: str
  description:
    description:
      - A description of what this stack set creates.
    type: str
  parameters:
    description:
      - A list of hashes of all the template variables for the stack. The value can be a string or a dict.
      - Dict can be used to set additional template parameter attributes like UsePreviousValue (see example).
    default: {}
    type: dict
  state:
    description:
      - If I(state=present), stack will be created.  If I(state=present) and if stack exists and template has changed, it will be updated.
        If I(state=absent), stack will be removed.
    default: present
    choices: [ present, absent ]
    type: str
  template:
    description:
      - The local path of the CloudFormation template.
      - This must be the full path to the file, relative to the working directory. If using roles this may look
        like C(roles/cloudformation/files/cloudformation-example.json).
      - If I(state=present) and the stack does not exist yet, either I(template), I(template_body) or I(template_url)
        must be specified (but only one of them).
      - If I(state=present), the stack does exist, and neither I(template), I(template_body) nor I(template_url)
        are specified, the previous template will be reused.
    type: path
  template_body:
    description:
      - Template body. Use this to pass in the actual body of the CloudFormation template.
      - If I(state=present) and the stack does not exist yet, either I(template), I(template_body) or I(template_url)
        must be specified (but only one of them).
      - If I(state=present), the stack does exist, and neither I(template), I(template_body) nor I(template_url)
        are specified, the previous template will be reused.
    type: str
  template_url:
    description:
      - Location of file containing the template body.
      - The URL must point to a template (max size 307,200 bytes) located in an S3 bucket in the same region
        as the stack.
      - If I(state=present) and the stack does not exist yet, either I(template), I(template_body) or I(template_url)
        must be specified (but only one of them).
      - If I(state=present), the stack does exist, and neither I(template), I(template_body) nor I(template_url)
        are specified, the previous template will be reused.
    type: str
  purge_stacks:
    description:
    - Only applicable when I(state=absent). Sets whether, when deleting a stack set, the stack instances should also be deleted.
    - By default, instances will be deleted. To keep stacks when stack set is deleted set I(purge_stacks=false).
    type: bool
    default: true
  wait:
    description:
    - Whether or not to wait for stack operation to complete. This includes waiting for stack instances to reach UPDATE_COMPLETE status.
    - If you choose not to wait, this module will not notify when stack operations fail because it will not wait for them to finish.
    type: bool
    default: false
  wait_timeout:
    description:
    - How long to wait (in seconds) for stacks to complete create/update/delete operations.
    default: 900
    type: int
  capabilities:
    description:
    - Capabilities allow stacks to create and modify IAM resources, which may include adding users or roles.
    - Currently the only available values are 'CAPABILITY_IAM' and 'CAPABILITY_NAMED_IAM'. Either or both may be provided.
    - >
        The following resources require that one or both of these parameters is specified: AWS::IAM::AccessKey,
        AWS::IAM::Group, AWS::IAM::InstanceProfile, AWS::IAM::Policy, AWS::IAM::Role, AWS::IAM::User, AWS::IAM::UserToGroupAddition
    type: list
    elements: str
    choices:
    - 'CAPABILITY_IAM'
    - 'CAPABILITY_NAMED_IAM'
  regions:
    description:
    - A list of AWS regions to create instances of a stack in. The I(region) parameter chooses where the Stack Set is created, and I(regions)
      specifies the region for stack instances.
    - At least one region must be specified to create a stack set. On updates, if fewer regions are specified only the specified regions will
      have their stack instances updated.
    type: list
    elements: str
  accounts:
    description:
    - A list of AWS accounts in which to create instance of CloudFormation stacks.
    - At least one region must be specified to create a stack set. On updates, if fewer regions are specified only the specified regions will
      have their stack instances updated.
    type: list
    elements: str
  administration_role_arn:
    description:
    - ARN of the administration role, meaning the role that CloudFormation Stack Sets use to assume the roles in your child accounts.
    - This defaults to C(arn:aws:iam::{{ account ID }}:role/AWSCloudFormationStackSetAdministrationRole) where C({{ account ID }}) is replaced with the
      account number of the current IAM role/user/STS credentials.
    aliases:
    - admin_role_arn
    - admin_role
    - administration_role
    type: str
  execution_role_name:
    description:
    - ARN of the execution role, meaning the role that CloudFormation Stack Sets assumes in your child accounts.
    - This MUST NOT be an ARN, and the roles must exist in each child account specified.
    - The default name for the execution role is C(AWSCloudFormationStackSetExecutionRole)
    aliases:
    - exec_role_name
    - exec_role
    - execution_role
    type: str
  tags:
    description:
      - Dictionary of tags to associate with stack and its resources during stack creation.
      - Can be updated later, updating tags removes previous entries.
    type: dict
  failure_tolerance:
    description:
    - Settings to change what is considered "failed" when running stack instance updates, and how many to do at a time.
    type: dict
    default: {}
    suboptions:
      fail_count:
        description:
        - The number of accounts, per region, for which this operation can fail before CloudFormation
          stops the operation in that region.
        - You must specify one of I(fail_count) and I(fail_percentage).
        type: int
      fail_percentage:
        type: int
        description:
        - The percentage of accounts, per region, for which this stack operation can fail before CloudFormation
          stops the operation in that region.
        - You must specify one of I(fail_count) and I(fail_percentage).
      parallel_percentage:
        type: int
        description:
        - The maximum percentage of accounts in which to perform this operation at one time.
        - You must specify one of I(parallel_count) and I(parallel_percentage).
        - Note that this setting lets you specify the maximum for operations.
          For large deployments, under certain circumstances the actual percentage may be lower.
      parallel_count:
        type: int
        description:
        - The maximum number of accounts in which to perform this operation at one time.
        - I(parallel_count) may be at most one more than the I(fail_count).
        - You must specify one of I(parallel_count) and I(parallel_percentage).
        - Note that this setting lets you specify the maximum for operations.
          For large deployments, under certain circumstances the actual count may be lower.

author:
  - "Ryan Scott Brown (@ryansb)"
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
- name: Create a stack set with instances in two accounts
  community.aws.cloudformation_stack_set:
    name: my-stack
    description: Test stack in two accounts
    state: present
    template_url: https://s3.amazonaws.com/my-bucket/cloudformation.template
    accounts:
      - 123456789012
      - 234567890123
    regions:
      - us-east-1

- name: on subsequent calls, templates are optional but parameters and tags can be altered
  community.aws.cloudformation_stack_set:
    name: my-stack
    state: present
    parameters:
      InstanceName: my_stacked_instance
    tags:
      foo: bar
      test: stack
    accounts:
      - 123456789012
      - 234567890123
    regions:
      - us-east-1

- name: The same type of update, but wait for the update to complete in all stacks
  community.aws.cloudformation_stack_set:
    name: my-stack
    state: present
    wait: true
    parameters:
      InstanceName: my_restacked_instance
    tags:
      foo: bar
      test: stack
    accounts:
      - 123456789012
      - 234567890123
    regions:
      - us-east-1

- name: Register new accounts (create new stack instances) with an existing stack set.
  community.aws.cloudformation_stack_set:
    name: my-stack
    state: present
    wait: true
    parameters:
      InstanceName: my_restacked_instance
    tags:
      foo: bar
      test: stack
    accounts:
      - 123456789012
      - 234567890123
      - 345678901234
    regions:
      - us-east-1
"""

RETURN = r"""
operations_log:
  type: list
  description: Most recent events in CloudFormation's event log. This may be from a previous run in some cases.
  returned: always
  sample:
  - action: CREATE
    creation_timestamp: '2018-06-18T17:40:46.372000+00:00'
    end_timestamp: '2018-06-18T17:41:24.560000+00:00'
    operation_id: Ansible-StackInstance-Create-0ff2af5b-251d-4fdb-8b89-1ee444eba8b8
    status: FAILED
    stack_instances:
    - account: '1234567890'
      region: us-east-1
      stack_set_id: TestStackPrime:19f3f684-aae9-4e67-ba36-e09f92cf5929
      status: OUTDATED
      status_reason: Account 1234567890 should have 'AWSCloudFormationStackSetAdministrationRole' role with trust relationship to CloudFormation service.

operations:
  description: All operations initiated by this run of the cloudformation_stack_set module
  returned: always
  type: list
  sample:
  - action: CREATE
    administration_role_arn: arn:aws:iam::1234567890:role/AWSCloudFormationStackSetAdministrationRole
    creation_timestamp: '2018-06-18T17:40:46.372000+00:00'
    end_timestamp: '2018-06-18T17:41:24.560000+00:00'
    execution_role_name: AWSCloudFormationStackSetExecutionRole
    operation_id: Ansible-StackInstance-Create-0ff2af5b-251d-4fdb-8b89-1ee444eba8b8
    operation_preferences:
      region_order:
      - us-east-1
      - us-east-2
    stack_set_id: TestStackPrime:19f3f684-aae9-4e67-ba36-e09f92cf5929
    status: FAILED
stack_instances:
  description: CloudFormation stack instances that are members of this stack set. This will also include their region and account ID.
  returned: state == present
  type: list
  sample:
    - account: '1234567890'
      region: us-east-1
      stack_set_id: TestStackPrime:19f3f684-aae9-4e67-ba36-e09f92cf5929
      status: OUTDATED
      status_reason: >
        Account 1234567890 should have 'AWSCloudFormationStackSetAdministrationRole' role with trust relationship to CloudFormation service.
    - account: '1234567890'
      region: us-east-2
      stack_set_id: TestStackPrime:19f3f684-aae9-4e67-ba36-e09f92cf5929
      status: OUTDATED
      status_reason: Cancelled since failure tolerance has exceeded
stack_set:
  type: dict
  description: Facts about the currently deployed stack set, its parameters, and its tags
  returned: state == present
  sample:
    administration_role_arn: arn:aws:iam::1234567890:role/AWSCloudFormationStackSetAdministrationRole
    capabilities: []
    description: test stack PRIME
    execution_role_name: AWSCloudFormationStackSetExecutionRole
    parameters: []
    stack_set_arn: arn:aws:cloudformation:us-east-1:1234567890:stackset/TestStackPrime:19f3f684-aae9-467-ba36-e09f92cf5929
    stack_set_id: TestStackPrime:19f3f684-aae9-4e67-ba36-e09f92cf5929
    stack_set_name: TestStackPrime
    status: ACTIVE
    tags:
      Some: Thing
      an: other
    template_body: |
      AWSTemplateFormatVersion: "2010-09-09"
      Parameters: {}
      Resources:
        Bukkit:
          Type: "AWS::S3::Bucket"
          Properties: {}
        other:
          Type: "AWS::SNS::Topic"
          Properties: {}
"""

import datetime
import itertools
import time
import uuid

try:
    from botocore.exceptions import BotoCoreError
    from botocore.exceptions import ClientError
except ImportError:
    # handled by AnsibleAWSModule
    pass

from ansible.module_utils._text import to_native
from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry
from ansible_collections.amazon.aws.plugins.module_utils.tagging import ansible_dict_to_boto3_tag_list
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict

from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule


def create_stack_set(module, stack_params, cfn):
    try:
        cfn.create_stack_set(aws_retry=True, **stack_params)
        return await_stack_set_exists(cfn, stack_params["StackSetName"])
    except (ClientError, BotoCoreError) as err:
        module.fail_json_aws(err, msg=f"Failed to create stack set {stack_params.get('StackSetName')}.")


def update_stack_set(module, stack_params, cfn):
    # if the state is present and the stack already exists, we try to update it.
    # AWS will tell us if the stack template and parameters are the same and
    # don't need to be updated.
    try:
        cfn.update_stack_set(**stack_params)
    except is_boto3_error_code("StackSetNotFound") as err:  # pylint: disable=duplicate-except
        module.fail_json_aws(err, msg="Failed to find stack set. Check the name & region.")
    except is_boto3_error_code("StackInstanceNotFound") as err:  # pylint: disable=duplicate-except
        module.fail_json_aws(
            err,
            msg=(
                "One or more stack instances were not found for this stack set. Double check "
                "the `accounts` and `regions` parameters."
            ),
        )
    except is_boto3_error_code("OperationInProgressException") as err:  # pylint: disable=duplicate-except
        module.fail_json_aws(
            err,
            msg=(
                "Another operation is already in progress on this stack set - please try again later. When making"
                " multiple cloudformation_stack_set calls, it's best to enable `wait: true` to avoid unfinished op"
                " errors."
            ),
        )
    except (ClientError, BotoCoreError) as err:  # pylint: disable=duplicate-except
        module.fail_json_aws(err, msg="Could not update stack set.")
    if module.params.get("wait"):
        await_stack_set_operation(
            module,
            cfn,
            operation_id=stack_params["OperationId"],
            stack_set_name=stack_params["StackSetName"],
            max_wait=module.params.get("wait_timeout"),
        )

    return True


def compare_stack_instances(cfn, stack_set_name, accounts, regions):
    instance_list = cfn.list_stack_instances(
        aws_retry=True,
        StackSetName=stack_set_name,
    )["Summaries"]
    desired_stack_instances = set(itertools.product(accounts, regions))
    existing_stack_instances = set((i["Account"], i["Region"]) for i in instance_list)
    # new stacks, existing stacks, unspecified stacks
    return (
        (desired_stack_instances - existing_stack_instances),
        existing_stack_instances,
        (existing_stack_instances - desired_stack_instances),
    )


@AWSRetry.jittered_backoff(retries=3, delay=4)
def stack_set_facts(cfn, stack_set_name):
    try:
        ss = cfn.describe_stack_set(StackSetName=stack_set_name)["StackSet"]
        ss["Tags"] = boto3_tag_list_to_ansible_dict(ss["Tags"])
        return ss
    except cfn.exceptions.from_code("StackSetNotFound"):
        # Return None if the stack doesn't exist
        return


def await_stack_set_operation(module, cfn, stack_set_name, operation_id, max_wait):
    wait_start = datetime.datetime.now()
    operation = None
    for i in range(max_wait // 15):
        try:
            operation = cfn.describe_stack_set_operation(StackSetName=stack_set_name, OperationId=operation_id)
            if operation["StackSetOperation"]["Status"] not in ("RUNNING", "STOPPING"):
                # Stack set has completed operation
                break
        except is_boto3_error_code("StackSetNotFound"):  # pylint: disable=duplicate-except
            pass
        except is_boto3_error_code("OperationNotFound"):  # pylint: disable=duplicate-except
            pass
        time.sleep(15)

    if operation and operation["StackSetOperation"]["Status"] not in ("FAILED", "STOPPED"):
        await_stack_instance_completion(
            module,
            cfn,
            stack_set_name=stack_set_name,
            # subtract however long we waited already
            max_wait=int(max_wait - (datetime.datetime.now() - wait_start).total_seconds()),
        )
    elif operation and operation["StackSetOperation"]["Status"] in ("FAILED", "STOPPED"):
        pass
    else:
        module.warn(
            f"Timed out waiting for operation {operation_id} on stack set {stack_set_name} after {max_wait} seconds."
            " Returning unfinished operation"
        )


def await_stack_instance_completion(module, cfn, stack_set_name, max_wait):
    to_await = None
    for i in range(max_wait // 15):
        try:
            stack_instances = cfn.list_stack_instances(StackSetName=stack_set_name)
            to_await = [inst for inst in stack_instances["Summaries"] if inst["Status"] != "CURRENT"]
            if not to_await:
                return stack_instances["Summaries"]
        except is_boto3_error_code("StackSetNotFound"):  # pylint: disable=duplicate-except
            # this means the deletion beat us, or the stack set is not yet propagated
            pass
        time.sleep(15)

    module.warn(
        f"Timed out waiting for stack set {stack_set_name} instances {', '.join(s['StackId'] for s in to_await)} to"
        f" complete after {max_wait} seconds. Returning unfinished operation"
    )


def await_stack_set_exists(cfn, stack_set_name):
    # AWSRetry will retry on `StackSetNotFound` errors for us
    ss = cfn.describe_stack_set(StackSetName=stack_set_name, aws_retry=True)["StackSet"]
    ss["Tags"] = boto3_tag_list_to_ansible_dict(ss["Tags"])
    return camel_dict_to_snake_dict(ss, ignore_list=("Tags",))


def describe_stack_tree(module, stack_set_name, operation_ids=None):
    jittered_backoff_decorator = AWSRetry.jittered_backoff(
        retries=5, delay=3, max_delay=5, catch_extra_error_codes=["StackSetNotFound"]
    )
    cfn = module.client("cloudformation", retry_decorator=jittered_backoff_decorator)
    result = dict()
    result["stack_set"] = camel_dict_to_snake_dict(
        cfn.describe_stack_set(
            StackSetName=stack_set_name,
            aws_retry=True,
        )["StackSet"]
    )
    result["stack_set"]["tags"] = boto3_tag_list_to_ansible_dict(result["stack_set"]["tags"])
    result["operations_log"] = sorted(
        camel_dict_to_snake_dict(
            cfn.list_stack_set_operations(
                StackSetName=stack_set_name,
                aws_retry=True,
            )
        )["summaries"],
        key=lambda x: x["creation_timestamp"],
    )
    result["stack_instances"] = sorted(
        [camel_dict_to_snake_dict(i) for i in cfn.list_stack_instances(StackSetName=stack_set_name)["Summaries"]],
        key=lambda i: i["region"] + i["account"],
    )

    if operation_ids:
        result["operations"] = []
        for op_id in operation_ids:
            try:
                result["operations"].append(
                    camel_dict_to_snake_dict(
                        cfn.describe_stack_set_operation(
                            StackSetName=stack_set_name,
                            OperationId=op_id,
                        )["StackSetOperation"]
                    )
                )
            except is_boto3_error_code("OperationNotFoundException"):  # pylint: disable=duplicate-except
                pass
    return result


def get_operation_preferences(module):
    params = dict()
    if module.params.get("regions"):
        params["RegionOrder"] = list(module.params["regions"])
    for param, api_name in {
        "fail_count": "FailureToleranceCount",
        "fail_percentage": "FailureTolerancePercentage",
        "parallel_percentage": "MaxConcurrentPercentage",
        "parallel_count": "MaxConcurrentCount",
    }.items():
        if module.params.get("failure_tolerance", {}).get(param):
            params[api_name] = module.params.get("failure_tolerance", {}).get(param)
    return params


def main():
    argument_spec = dict(
        name=dict(required=True),
        description=dict(),
        wait=dict(type="bool", default=False),
        wait_timeout=dict(type="int", default=900),
        state=dict(default="present", choices=["present", "absent"]),
        purge_stacks=dict(type="bool", default=True),
        parameters=dict(type="dict", default={}),
        template=dict(type="path"),
        template_url=dict(),
        template_body=dict(),
        capabilities=dict(type="list", elements="str", choices=["CAPABILITY_IAM", "CAPABILITY_NAMED_IAM"]),
        regions=dict(type="list", elements="str"),
        accounts=dict(type="list", elements="str"),
        failure_tolerance=dict(
            type="dict",
            default={},
            options=dict(
                fail_count=dict(type="int"),
                fail_percentage=dict(type="int"),
                parallel_percentage=dict(type="int"),
                parallel_count=dict(type="int"),
            ),
            mutually_exclusive=[
                ["fail_count", "fail_percentage"],
                ["parallel_count", "parallel_percentage"],
            ],
        ),
        administration_role_arn=dict(aliases=["admin_role_arn", "administration_role", "admin_role"]),
        execution_role_name=dict(aliases=["execution_role", "exec_role", "exec_role_name"]),
        tags=dict(type="dict"),
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        mutually_exclusive=[["template_url", "template", "template_body"]],
        supports_check_mode=True,
    )

    # Wrap the cloudformation client methods that this module uses with
    # automatic backoff / retry for throttling error codes
    jittered_backoff_decorator = AWSRetry.jittered_backoff(
        retries=10, delay=3, max_delay=30, catch_extra_error_codes=["StackSetNotFound"]
    )
    cfn = module.client("cloudformation", retry_decorator=jittered_backoff_decorator)
    existing_stack_set = stack_set_facts(cfn, module.params["name"])

    operation_uuid = to_native(uuid.uuid4())
    operation_ids = []
    # collect the parameters that are passed to boto3. Keeps us from having so many scalars floating around.
    stack_params = {}
    state = module.params["state"]
    if state == "present" and not module.params["accounts"]:
        module.fail_json(
            msg=(
                "Can't create a stack set without choosing at least one account. "
                "To get the ID of the current account, use the aws_caller_info module."
            )
        )

    module.params["accounts"] = [to_native(a) for a in module.params["accounts"]]

    stack_params["StackSetName"] = module.params["name"]
    if module.params.get("description"):
        stack_params["Description"] = module.params["description"]

    if module.params.get("capabilities"):
        stack_params["Capabilities"] = module.params["capabilities"]

    if module.params["template"] is not None:
        with open(module.params["template"], "r") as tpl:
            stack_params["TemplateBody"] = tpl.read()
    elif module.params["template_body"] is not None:
        stack_params["TemplateBody"] = module.params["template_body"]
    elif module.params["template_url"] is not None:
        stack_params["TemplateURL"] = module.params["template_url"]
    else:
        # no template is provided, but if the stack set exists already, we can use the existing one.
        if existing_stack_set:
            stack_params["UsePreviousTemplate"] = True
        else:
            module.fail_json(
                msg=(
                    f"The Stack Set {module.params['name']} does not exist, and no template was provided. Provide one"
                    " of `template`, `template_body`, or `template_url`"
                )
            )

    stack_params["Parameters"] = []
    for k, v in module.params["parameters"].items():
        if isinstance(v, dict):
            # set parameter based on a dict to allow additional CFN Parameter Attributes
            param = dict(ParameterKey=k)

            if "value" in v:
                param["ParameterValue"] = to_native(v["value"])

            if "use_previous_value" in v and bool(v["use_previous_value"]):
                param["UsePreviousValue"] = True
                param.pop("ParameterValue", None)

            stack_params["Parameters"].append(param)
        else:
            # allow default k/v configuration to set a template parameter
            stack_params["Parameters"].append({"ParameterKey": k, "ParameterValue": str(v)})

    if module.params.get("tags") and isinstance(module.params.get("tags"), dict):
        stack_params["Tags"] = ansible_dict_to_boto3_tag_list(module.params["tags"])

    if module.params.get("administration_role_arn"):
        # TODO loosen the semantics here to autodetect the account ID and build the ARN
        stack_params["AdministrationRoleARN"] = module.params["administration_role_arn"]
    if module.params.get("execution_role_name"):
        stack_params["ExecutionRoleName"] = module.params["execution_role_name"]

    result = {}

    if module.check_mode:
        if state == "absent" and existing_stack_set:
            module.exit_json(changed=True, msg="Stack set would be deleted", meta=[])
        elif state == "absent" and not existing_stack_set:
            module.exit_json(changed=False, msg="Stack set doesn't exist", meta=[])
        elif state == "present" and not existing_stack_set:
            module.exit_json(changed=True, msg="New stack set would be created", meta=[])
        elif state == "present" and existing_stack_set:
            new_stacks, existing_stacks, unspecified_stacks = compare_stack_instances(
                cfn,
                module.params["name"],
                module.params["accounts"],
                module.params["regions"],
            )
            if new_stacks:
                module.exit_json(changed=True, msg="New stack instance(s) would be created", meta=[])
            elif unspecified_stacks and module.params.get("purge_stack_instances"):
                module.exit_json(changed=True, msg="Old stack instance(s) would be deleted", meta=[])
        else:
            # TODO: need to check the template and other settings for correct check mode
            module.exit_json(changed=False, msg="No changes detected", meta=[])

    changed = False
    if state == "present":
        if not existing_stack_set:
            # on create this parameter has a different name, and cannot be referenced later in the job log
            stack_params["ClientRequestToken"] = f"Ansible-StackSet-Create-{operation_uuid}"
            changed = True
            create_stack_set(module, stack_params, cfn)
        else:
            stack_params["OperationId"] = f"Ansible-StackSet-Update-{operation_uuid}"
            operation_ids.append(stack_params["OperationId"])
            if module.params.get("regions"):
                stack_params["OperationPreferences"] = get_operation_preferences(module)
            changed |= update_stack_set(module, stack_params, cfn)

            await_stack_set_operation(
                module,
                cfn,
                operation_id=stack_params["OperationId"],
                stack_set_name=stack_params["StackSetName"],
                max_wait=module.params.get("wait_timeout"),
            )

        # now create/update any appropriate stack instances
        new_stack_instances, existing_stack_instances, unspecified_stack_instances = compare_stack_instances(
            cfn,
            module.params["name"],
            module.params["accounts"],
            module.params["regions"],
        )
        if new_stack_instances:
            operation_ids.append(f"Ansible-StackInstance-Create-{operation_uuid}")
            changed = True
            cfn.create_stack_instances(
                StackSetName=module.params["name"],
                Accounts=list(set(acct for acct, region in new_stack_instances)),
                Regions=list(set(region for acct, region in new_stack_instances)),
                OperationPreferences=get_operation_preferences(module),
                OperationId=operation_ids[-1],
            )
        else:
            operation_ids.append(f"Ansible-StackInstance-Update-{operation_uuid}")
            cfn.update_stack_instances(
                StackSetName=module.params["name"],
                Accounts=list(set(acct for acct, region in existing_stack_instances)),
                Regions=list(set(region for acct, region in existing_stack_instances)),
                OperationPreferences=get_operation_preferences(module),
                OperationId=operation_ids[-1],
            )
        for op in operation_ids:
            await_stack_set_operation(
                module,
                cfn,
                operation_id=op,
                stack_set_name=module.params["name"],
                max_wait=module.params.get("wait_timeout"),
            )

    elif state == "absent":
        if not existing_stack_set:
            module.exit_json(msg=f"Stack set {module.params['name']} does not exist")
        if module.params.get("purge_stack_instances") is False:
            pass
        try:
            cfn.delete_stack_set(
                StackSetName=module.params["name"],
            )
            module.exit_json(msg=f"Stack set {module.params['name']} deleted")
        except is_boto3_error_code("OperationInProgressException") as e:  # pylint: disable=duplicate-except
            module.fail_json_aws(
                e, msg=f"Cannot delete stack {module.params['name']} while there is an operation in progress"
            )
        except is_boto3_error_code("StackSetNotEmptyException"):  # pylint: disable=duplicate-except
            delete_instances_op = f"Ansible-StackInstance-Delete-{operation_uuid}"
            cfn.delete_stack_instances(
                StackSetName=module.params["name"],
                Accounts=module.params["accounts"],
                Regions=module.params["regions"],
                RetainStacks=(not module.params.get("purge_stacks")),
                OperationId=delete_instances_op,
            )
            await_stack_set_operation(
                module,
                cfn,
                operation_id=delete_instances_op,
                stack_set_name=stack_params["StackSetName"],
                max_wait=module.params.get("wait_timeout"),
            )
            try:
                cfn.delete_stack_set(
                    StackSetName=module.params["name"],
                )
            except is_boto3_error_code("StackSetNotEmptyException") as exc:  # pylint: disable=duplicate-except
                # this time, it is likely that either the delete failed or there are more stacks.
                instances = cfn.list_stack_instances(
                    StackSetName=module.params["name"],
                )
                stack_states = ", ".join(
                    "(account={Account}, region={Region}, state={Status})".format(**i) for i in instances["Summaries"]
                )
                module.fail_json_aws(
                    exc,
                    msg="Could not purge all stacks, or not all accounts/regions were chosen for deletion: "
                    + stack_states,
                )
            module.exit_json(changed=True, msg=f"Stack set {module.params['name']} deleted")

    result.update(**describe_stack_tree(module, stack_params["StackSetName"], operation_ids=operation_ids))
    if any(o["status"] == "FAILED" for o in result["operations"]):
        module.fail_json(msg="One or more operations failed to execute", **result)
    module.exit_json(changed=changed, **result)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2013, André Paramés <git@andreparames.com>
# Based on the Git module by Michael DeHaan <michael.dehaan@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: bzr
author:
  - André Paramés (@andreparames)
short_description: Deploy software (or files) from bzr branches
description:
  - Manage C(bzr) branches to deploy files or software.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  name:
    description:
      - SSH or HTTP protocol address of the parent branch.
    aliases: [parent]
    required: true
    type: str
  dest:
    description:
      - Absolute path of where the branch should be cloned to.
    required: true
    type: path
  version:
    description:
      - What version of the branch to clone. This can be the bzr revno or revid.
    default: head
    type: str
  force:
    description:
      - If V(true), any modified files in the working tree is discarded.
    type: bool
    default: false
  executable:
    description:
      - Path to C(bzr) executable to use. If not supplied, the normal mechanism for resolving binary paths is used.
    type: str
"""

EXAMPLES = r"""
- name: Checkout
  community.general.bzr:
    name: bzr+ssh://foosball.example.org/path/to/branch
    dest: /srv/checkout
    version: 22
"""

import os
import re

from ansible.module_utils.basic import AnsibleModule


class Bzr:
    def __init__(self, module, parent, dest, version, bzr_path):
        self.module = module
        self.parent = parent
        self.dest = dest
        self.version = version
        self.bzr_path = bzr_path

    def _command(self, args_list, cwd=None, **kwargs):
        (rc, out, err) = self.module.run_command([self.bzr_path] + args_list, cwd=cwd, **kwargs)
        return (rc, out, err)

    def get_version(self):
        """samples the version of the bzr branch"""

        cmd = [self.bzr_path, "revno"]
        rc, stdout, stderr = self.module.run_command(cmd, cwd=self.dest)
        revno = stdout.strip()
        return revno

    def clone(self):
        """makes a new bzr branch if it does not already exist"""
        dest_dirname = os.path.dirname(self.dest)
        try:
            os.makedirs(dest_dirname)
        except Exception:
            pass
        if self.version.lower() != "head":
            args_list = ["branch", "-r", self.version, self.parent, self.dest]
        else:
            args_list = ["branch", self.parent, self.dest]
        return self._command(args_list, check_rc=True, cwd=dest_dirname)

    def has_local_mods(self):
        cmd = [self.bzr_path, "status", "-S"]
        rc, stdout, stderr = self.module.run_command(cmd, cwd=self.dest)
        lines = stdout.splitlines()
        mods_re = re.compile("^\\?\\?.*$")

        lines = [c for c in lines if not mods_re.search(c)]
        return len(lines) > 0

    def reset(self, force):
        """
        Resets the index and working tree to head.
        Discards any changes to tracked files in the working
        tree since that commit.
        """
        if not force and self.has_local_mods():
            self.module.fail_json(msg="Local modifications exist in branch (force=false).")
        return self._command(["revert"], check_rc=True, cwd=self.dest)

    def fetch(self):
        """updates branch from remote sources"""
        if self.version.lower() != "head":
            (rc, out, err) = self._command(["pull", "-r", self.version], cwd=self.dest)
        else:
            (rc, out, err) = self._command(["pull"], cwd=self.dest)
        if rc != 0:
            self.module.fail_json(msg="Failed to pull")
        return (rc, out, err)

    def switch_version(self):
        """once pulled, switch to a particular revno or revid"""
        if self.version.lower() != "head":
            args_list = ["revert", "-r", self.version]
        else:
            args_list = ["revert"]
        return self._command(args_list, check_rc=True, cwd=self.dest)


# ===========================================


def main():
    module = AnsibleModule(
        argument_spec=dict(
            dest=dict(type="path", required=True),
            name=dict(type="str", required=True, aliases=["parent"]),
            version=dict(type="str", default="head"),
            force=dict(type="bool", default=False),
            executable=dict(type="str"),
        )
    )
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}

    dest = module.params["dest"]
    parent = module.params["name"]
    version = module.params["version"]
    force = module.params["force"]
    bzr_path = module.params["executable"] or module.get_bin_path("bzr", True)

    bzrconfig = os.path.join(dest, ".bzr", "branch", "branch.conf")

    rc, out, err = (0, None, None)

    bzr = Bzr(module, parent, dest, version, bzr_path)

    # if there is no bzr configuration, do a branch operation
    # else pull and switch the version
    before = None
    local_mods = False
    if not os.path.exists(bzrconfig):
        (rc, out, err) = bzr.clone()

    else:
        # else do a pull
        local_mods = bzr.has_local_mods()
        before = bzr.get_version()
        (rc, out, err) = bzr.reset(force)
        if rc != 0:
            module.fail_json(msg=err)
        (rc, out, err) = bzr.fetch()
        if rc != 0:
            module.fail_json(msg=err)

    # switch to version specified regardless of whether
    # we cloned or pulled
    (rc, out, err) = bzr.switch_version()

    # determine if we changed anything
    after = bzr.get_version()
    changed = False

    if before != after or local_mods:
        changed = True

    module.exit_json(changed=changed, before=before, after=after)


if __name__ == "__main__":
    main()
